"""
Buddy conversation
"""

import datetime
import logging
import urllib.parse

from typing import List, Tuple, TYPE_CHECKING

import nuqql.win

from .conversation import Conversation

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from nuqql.account import Account  # noqa
    from nuqql.backend import Backend  # noqa
    from nuqql.buddy import Buddy   # noqa

logger = logging.getLogger(__name__)


class BuddyConversation(Conversation):
    """
    Class for conversations with buddies
    """

    def __init__(self, backend: "Backend", account: "Account",
                 name: str) -> None:
        Conversation.__init__(self, backend, account, name)

        # conv already in backend or new (temporary) conv in nuqql only
        self.temporary = False

        self.peers: List["Buddy"] = []
        self.wins.list_win = nuqql.win.MAIN_WINS["list"]
        self.history.init_logger()

    def create_windows(self) -> None:
        """
        Create windows for this conversation
        """

        logger.debug("creating windows for conversation %s", self.name)

        # create standard chat windows
        log_title = "Chat log with {0}".format(self.name)
        input_title = "Message to {0}".format(self.name)
        self._create_windows_common(log_title, input_title)

        # try to read old messages from message history
        self.history.init_log_from_file()

        # if this conversation belongs to a group chat invite, display special
        # event to the user
        if self.peers:
            buddy = self.peers[0]
            if buddy.status == "grp_invite":
                msg = "<You are invited to this group chat. " \
                        "Enter \"/join\" to accept or \"/part\" to decline " \
                        "this invite.>"
                self.log("<event>", msg, own=True)

    def get_name(self) -> str:
        """
        Get the name of the conversation, depending on type
        """

        notify = self._get_name_notification()
        if self.peers:
            peer = self.peers[0]
            # unquote the alias, e.g., remove %20 for whitespace
            alias = urllib.parse.unquote(peer.alias)
            return "{0}[{1}] {2}".format(notify, peer.status, alias)
        return "{0}[{1}] {2}".format(notify, "off", self.name)

    def get_key(self) -> Tuple:
        """
        Get a key for sorting this conversation
        """

        # defaults
        sort_notify = 0 - self.notification
        sort_used = 0.0
        sort_type = 0
        sort_status = 0
        sort_name = self.name

        # get sort key from config
        config = nuqql.config.get("conversations")
        sort_key = config["sort_key"]

        if self.peers:
            peer = self.peers[0]
            if peer.status != "off" and sort_key in self.stats:
                sort_used = 0 - self.stats[sort_key]
            try:
                sort_status = self.status_key[peer.status]
            except KeyError:
                sort_status = len(self.status_key) + 1
            sort_name = peer.alias

        # return tuple of sort keys
        return sort_notify, sort_used, sort_type, sort_status, sort_name

    def group_send_msg(self, msg: str) -> None:
        """
        Send message to a group conversation coming from the UI/input window
        """

        assert self.account and self.backend and self.backend.client
        logger.debug("sending group message %s in conversation %s", msg,
                     self.name)

        # check for special commands
        if msg == "/names":
            self.log("you", msg, own=True)

            # TODO: use peers list for this?
            # create user list command
            msg = "account {} chat users {}".format(self.account.aid,
                                                    self.name)
            # send command message to backend
            self.backend.client.send_command(msg)
            return

        if msg == "/part":
            self.log("you", msg, own=True)

            # create chat part command
            msg = "account {} chat part {}".format(self.account.aid,
                                                   self.name)
            # send command message to backend
            self.backend.client.send_command(msg)
            return

        if msg.startswith("/invite "):
            self.log("you", msg, own=True)

            parts = msg.split()
            if len(parts) > 1:
                # create chat invite command
                user = parts[1]
                msg = "account {} chat invite {} {}".format(
                    self.account.aid, self.name, user)
                # send command message to backend
                self.backend.client.send_command(msg)
                return

        if msg == "/join":
            self.log("you", msg, own=True)

            # TODO: allow specification of another group chat?
            # create chat join command
            msg = "account {} chat join {}".format(self.account.aid,
                                                   self.name)
            # send command message to backend
            self.backend.client.send_command(msg)
            return

        # send group chat message
        # note: backends send all group chat messages including our own to us,
        # so do not log them here
        self.backend.client.send_group_msg(self.account.aid, self.name, msg)

    def send_msg(self, msg: str) -> None:
        """
        Send message coming from the UI/input window
        """

        logger.debug("sending message %s in conversation %s", msg, self.name)

        if not self.account or not self.backend or not self.backend.client:
            return

        # statistics
        self.stats["last_send"] = datetime.datetime.now().timestamp()
        self.stats["num_send"] += 1

        # redraw list_win in case sorting is affected by stats update above
        self.wins.list_win.redraw_pad()

        # handle group chat messages separately
        if self.peers:
            buddy = self.peers[0]
            if buddy.status == "grp" or buddy.status == "grp_invite":
                self.group_send_msg(msg)
                return

        # send message and log it in the history file
        self.backend.client.send_msg(self.account.aid, self.name, msg)
        log_msg = self.log("you", msg, own=True)
        self.history.log_to_file(log_msg)

    def set_lastread(self) -> None:
        """
        Helper that sets lastread to the last message in the conversation,
        thus, marking all messages as read.
        """

        log_msg = None
        if self.wins.log_win.list:
            log_msg = self.wins.log_win.list[-1]
            # do not put new conversation event in last_read
            if log_msg.sender == "<event>":
                log_msg = None
                if len(self.wins.log_win.list) > 1:
                    log_msg = self.wins.log_win.list[-2]

        # if there is a log message, write it to lastread
        if log_msg:
            logger.debug("setting lastread in conversation %s to %s",
                         self.name, log_msg)
            self.history.set_lastread(log_msg)



"""
Nuqql message parsing
"""

import time
import html
import re

from typing import Tuple


def parse_error_msg(orig_msg: str) -> Tuple:
    """
    Parse "error" message received from backend

    Format:
        "error: %s\r\n"
    """

    error = orig_msg[7:]
    return "error", error


def parse_info_msg(orig_msg: str) -> Tuple:
    """
    Parse "info" message received from backend

    Format:
        "info: %s\r\n"
    """

    info = orig_msg[6:]

    # filter known messages that would spam the log. TODO: change this and/or
    # other message formats/protocol behaviour
    if info.startswith("got buddies for account "):
        # filter got buddies info spam
        return ("", )

    return "info", info


def parse_account_msg(orig_msg: str) -> Tuple:
    """
    Parse "account" message received from backend

    Format:
        "account: %d %s %s %s [%s]\r\n"
    """

    orig_msg = orig_msg[9:]
    part = orig_msg.split(" ")
    acc_id = part[0]
    acc_alias = part[1]
    acc_prot = part[2].lower()
    acc_user = part[3]
    acc_status = part[4]    # ignore [ and ] for now
    return "account", acc_id, acc_alias, acc_prot, acc_user, acc_status


def parse_collect_msg(orig_msg: str) -> Tuple:
    """
    Parse "collect" message received from backend
    """

    # collect response and message have the same message format
    return parse_message_msg(orig_msg)


def parse_message_msg(orig_msg: str) -> Tuple:
    """
    Parse "message" message received from backend
    """

    orig_msg = orig_msg[9:]
    part = orig_msg.split(" ")
    acc = part[0]
    acc_name = part[1]
    tstamp = part[2]
    sender = part[3]
    msg = " ".join(part[4:])
    msg = "\n".join(re.split("<br/?>", msg, flags=re.IGNORECASE))
    msg = html.unescape(msg)

    return "message", acc, acc_name, int(tstamp), sender, msg


def parse_buddy_msg(orig_msg: str) -> Tuple:
    """
    Parse "buddy" message received from backend
    """

    orig_msg = orig_msg[7:]
    # <acc> status: <Offline/Available> name: <name> alias: <alias>
    part = orig_msg.split(" ")
    acc = part[0]
    status = part[2]
    name = part[4]
    alias = part[6]
    return "buddy", acc, status, name, alias


def parse_status_msg(orig_msg: str) -> Tuple:
    """
    Parse "status" message received from backend
    """

    orig_msg = orig_msg[8:]
    # account <acc> status: <status>
    part = orig_msg.split(" ")
    acc = part[1]
    status = part[3]

    return "status", acc, status


def parse_chat_msg(orig_msg: str) -> Tuple:
    """
    Parse "chat" message received from backend
    """

    orig_msg = orig_msg[6:]
    # list: <acc> <chat> <chat_alias> <nick>
    # user: <acc> <chat> <user> <user_alias> <status>
    part = orig_msg.split(" ")
    if len(part) < 5:
        # TODO: return a parsing error or something similar?
        return ("", )

    # common entries
    ctype = part[0]
    acc = part[1]
    chat = part[2]

    # list message
    if ctype == "list:":
        chat_alias = part[3]
        nick = part[4]
        return "chat", ctype, acc, chat, chat_alias, nick

    # user message
    if ctype == "user:" and len(part) >= 6:
        user = part[3]
        user_alias = part[4]
        status = part[5]
        return "chat", ctype, acc, chat, user, user_alias, status

    # msg message
    # TODO: remove duplicate code?
    if ctype == "msg:" and len(part) >= 6:
        tstamp = part[3]
        sender = part[4]
        msg = " ".join(part[5:])
        msg = "\n".join(re.split("<br/?>", msg, flags=re.IGNORECASE))
        msg = html.unescape(msg)
        return "chat", ctype, acc, chat, int(tstamp), sender, msg

    return ("", )


# dictionary for parsing functions, used by parse_msg()
PARSE_FUNCTIONS = {
    "message:": parse_message_msg,
    "collect:": parse_collect_msg,
    "buddy:": parse_buddy_msg,
    "account:": parse_account_msg,
    "status:": parse_status_msg,
    "chat:": parse_chat_msg,
    "info:": parse_info_msg,
    "error:": parse_error_msg,
}


def parse_msg(orig_msg: str) -> Tuple:
    """
    Parse message received from backend,
    calls more specific parsing functions
    """

    # extract message type and then call respectice parsing function
    msg_type = orig_msg.split(maxsplit=1)[0]
    try:
        return PARSE_FUNCTIONS[msg_type](orig_msg)
    except KeyError:
        # return this as parsing error
        acc = "-1"
        acc_name = "error"
        tstamp = int(time.time())
        sender = "<backend>"
        msg = "Error parsing message: " + orig_msg
        return "parsing error", acc, acc_name, tstamp, sender, msg



"""
Nuqql Buddy
"""

import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from nuqql.backend import Backend  # noqa
    from nuqql.account import Account  # noqa

logger = logging.getLogger(__name__)


class Buddy:
    """
    Class for Buddies
    """

    def __init__(self, backend: "Backend", account: "Account",
                 name: "str") -> None:
        self.backend = backend
        self.account = account
        self.name = name
        self.alias = name
        self.status = "off"     # use short status name
        self.updated = True
        logger.debug("created buddy: backend %s, account %s, "
                     "name %s, alias %s, status %s",
                     self.backend.name, self.account.aid, self.name,
                     self.alias, self.status)

    # dictionary for mapping status names to shorter version (key: lower case)
    status_map = {
        "offline": "off",
        "available": "on",
        "away": "afk",
        "group_chat": "grp",
        "group_chat_invite": "grp_invite",
    }

    def set_status(self, status: str) -> None:
        """
        Set status of buddy; convert status to something shorter
        """

        try:
            self.status = Buddy.status_map[status.lower()]
        except KeyError:
            self.status = status

    def update(self, status: str, alias: str) -> bool:
        """
        Update Buddy
        """

        # save old status and alias to check if buddy has changed
        old_status = self.status
        old_alias = self.alias

        # set new values
        self.updated = True
        self.set_status(status)
        self.alias = alias

        # check if buddy has changed
        if old_status != self.status or old_alias != self.alias:
            logger.debug("updated buddy: backend %s, account %s, "
                         "name %s, alias %s, status %s",
                         self.backend.name, self.account.aid, self.name,
                         self.alias, self.status)
            return True

        # has not changed
        return False



"""
Nuqql UI Input Windows
"""

import curses
import logging
import unicodedata

from typing import TYPE_CHECKING, Any

from .win import Win, MAIN_WINS

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from nuqql.config import WinConfig  # noqa
    from nuqql.conversation import Conversation  # noqa

logger = logging.getLogger(__name__)


class InputWin(Win):
    """
    Class for Input Windows
    """

    def __init__(self, config: "WinConfig", conversation: "Conversation",
                 title: str) -> None:
        Win.__init__(self, config, conversation, title)

        # input message
        self.msg = ""

    def redraw_pad(self) -> None:
        # if terminal size is invalid, stop here
        if not self.config.is_terminal_valid():
            return

        pos_y, pos_x = self.config.get_pos()
        win_size_y, win_size_x = self.config.get_size()

        self._move_pad()
        self._check_borders()
        self.pad.refresh(self.state.pad_y, self.state.pad_x,
                         pos_y + 1, pos_x + 1,
                         pos_y + win_size_y - 2,
                         pos_x + win_size_x - 2)

    def _cursor_up(self, *args: Any) -> None:
        segment = args[0]
        if self.state.cur_y > 0:
            logger.debug("moving cursor up")
            self.pad.move(self.state.cur_y - 1,
                          min(self.state.cur_x,
                              len(segment[self.state.cur_y - 1])))

        # display changes in the pad
        self.redraw_pad()

    def _cursor_down(self, *args: Any) -> None:
        # pad properties
        pad_y_max, unused_pad_x_max = self.pad.getmaxyx()

        segment = args[0]
        if self.state.cur_y < pad_y_max and \
           self.state.cur_y < len(segment) - 1:
            logger.debug("moving cursor down")
            self.pad.move(self.state.cur_y + 1,
                          min(self.state.cur_x,
                              len(segment[self.state.cur_y + 1])))

        # display changes in the pad
        self.redraw_pad()

    def _cursor_left(self, *args: Any) -> None:
        if self.state.cur_x > 0:
            logger.debug("moving cursor left")
            self.pad.move(self.state.cur_y, self.state.cur_x - 1)

        # display changes in the pad
        self.redraw_pad()

    def _cursor_right(self, *args: Any) -> None:
        # pad properties
        unused_pad_y_max, pad_x_max = self.pad.getmaxyx()

        segment = args[0]
        if self.state.cur_x < pad_x_max and \
           self.state.cur_x < len(segment[self.state.cur_y]):
            logger.debug("moving cursor right")
            self.pad.move(self.state.cur_y, self.state.cur_x + 1)

        # display changes in the pad
        self.redraw_pad()

    def _cursor_line_start(self, *args: Any) -> None:
        if self.state.cur_x > 0:
            logger.debug("moving cursor to start of line")
            self.pad.move(self.state.cur_y, 0)

        # display changes in the pad
        self.redraw_pad()

    def _cursor_line_end(self, *args: Any) -> None:
        # pad properties
        unused_pad_y_max, pad_x_max = self.pad.getmaxyx()

        segment = args[0]
        if self.state.cur_x < pad_x_max and \
           self.state.cur_x < len(segment[self.state.cur_y]):
            logger.debug("moving cursor to end of line")
            self.pad.move(self.state.cur_y, len(segment[self.state.cur_y]))

        # display changes in the pad
        self.redraw_pad()

    def _cursor_msg_start(self, *args: Any) -> None:
        if self.state.cur_y > 0 or self.state.cur_x > 0:
            logger.debug("moving cursor to start of message")
            self.pad.move(0, 0)

        # display changes in the pad
        self.redraw_pad()

    def _cursor_msg_end(self, *args: Any) -> None:
        segment = args[0]
        if self.state.cur_y < len(segment) - 1 or \
           self.state.cur_x < len(segment[-1]):
            logger.debug("moving cursor to end of message")
            self.pad.move(len(segment) - 1, len(segment[-1]))

        # display changes in the pad
        self.redraw_pad()

    def _send_msg(self, *args: Any) -> None:
        # do not send empty messages
        if self.msg == "":
            return

        logger.debug("sending message %s", self.msg)

        # let conversation actually send the message
        self.conversation.send_msg(self.msg)

        # reset input
        self.msg = ""
        self.pad.erase()

        # reset pad size
        win_size_y, win_size_x = self.win.getmaxyx()
        self.pad.resize(win_size_y - 2, win_size_x - 2)

        # display changes in the pad
        self.redraw_pad()

    def _delete_char(self, *args: Any) -> None:
        segment = args[0]
        if self.state.cur_x > 0:
            # delete charater within a line
            logger.debug("deleting character")
            segment[self.state.cur_y] = \
                segment[self.state.cur_y][:self.state.cur_x - 1] +\
                segment[self.state.cur_y][self.state.cur_x:]
        elif self.state.cur_y > 0:
            # delete newline
            logger.debug("deleting newline character")
            old_prev_len = len(segment[self.state.cur_y - 1])
            segment[self.state.cur_y - 1] = segment[self.state.cur_y - 1] +\
                segment[self.state.cur_y]
            segment = segment[:self.state.cur_y] + \
                segment[self.state.cur_y + 1:]
            # resize pad if concatenated line is longer than pad
            pad_size_y, pad_size_x = self.pad.getmaxyx()
            if len(segment[self.state.cur_y - 1]) + 2 > pad_size_x:
                pad_size_x = len(segment[self.state.cur_y - 1]) + 2
                self.pad.resize(pad_size_y, pad_size_x)
        else:
            # at top left, do nothing
            return

        # reconstruct and display message
        self.msg = "\n".join(segment)
        self.pad.erase()
        self.pad.addstr(self.msg)
        # move cursor to new position
        if self.state.cur_x > 0:
            self.pad.move(self.state.cur_y, self.state.cur_x - 1)
        elif self.state.cur_y > 0:
            self.pad.move(self.state.cur_y - 1, old_prev_len)

        # display changes in the pad
        self.redraw_pad()

    def _delete_char_right(self, *args: Any) -> None:
        # argument is segemented message, i.e. lines of the message
        segment = args[0]

        # delete character
        if self.state.cur_x < len(segment[self.state.cur_y]):
            # delete charater within a line
            logger.debug("deleting right character")
            segment[self.state.cur_y] = \
                segment[self.state.cur_y][:self.state.cur_x] +\
                segment[self.state.cur_y][self.state.cur_x + 1:]
        elif self.state.cur_y < len(segment) - 1:
            # delete newline
            logger.debug("deleting right newline character")
            segment[self.state.cur_y] = segment[self.state.cur_y] +\
                segment[self.state.cur_y + 1]
            del segment[self.state.cur_y + 1]
            # resize pad if concatenated line is longer than pad
            pad_size_y, pad_size_x = self.pad.getmaxyx()
            if len(segment[self.state.cur_y]) + 2 > pad_size_x:
                pad_size_x = len(segment[self.state.cur_y]) + 2
                self.pad.resize(pad_size_y, pad_size_x)
        else:
            # at bottom right, do nothing
            return

        # reconstruct and display message
        self.msg = "\n".join(segment)
        self.pad.erase()
        self.pad.addstr(self.msg)

        # move cursor to new position
        self.pad.move(self.state.cur_y, self.state.cur_x)

        # display changes in the pad
        self.redraw_pad()

    def _delete_line_end(self, *args: Any) -> None:
        segment = args[0]

        # delete from cursor to end of line
        logger.debug("deleting from cursor to end of line")
        segment[self.state.cur_y] = \
            segment[self.state.cur_y][:self.state.cur_x]

        # reconstruct message
        self.msg = "\n".join(segment)
        self.pad.erase()
        self.pad.addstr(self.msg)

        # display changes in the pad
        self.redraw_pad()

    def _delete_line(self, *args: Any) -> None:
        segment = args[0]

        # delete the current line
        logger.debug("deleting current line")
        del segment[self.state.cur_y]

        # reconstruct message
        self.msg = "\n".join(segment)
        self.pad.erase()
        self.pad.addstr(self.msg)

        # move cursor to new position
        if len(segment) <= self.state.cur_y:
            self.state.cur_y = max(0, len(segment) - 1)
        if not segment:
            self.state.cur_x = 0
        elif len(segment[self.state.cur_y]) < self.state.cur_x:
            self.state.cur_x = len(segment[self.state.cur_y])
        self.pad.move(self.state.cur_y, self.state.cur_x)

        # display changes in the pad
        self.redraw_pad()

    def go_back(self, *_args: Any) -> None:
        """
        Close conversation windows
        """

        logger.debug("leaving input window")
        self.state.active = False
        self.conversation.wins.log_win.state.active = False

        # disable cursor
        curses.curs_set(0)

        # redraw main windows
        MAIN_WINS["log"].redraw()

        # assume user read all messages and set lastread to last message
        self.conversation.set_lastread()

    def _go_log(self, *args: Any) -> None:
        """
        Jump to log
        """

        logger.debug("jumping to log")
        self.state.active = False
        self.conversation.wins.log_win.state.active = True
        self.conversation.wins.log_win.zoom_win()

    def _go_log_search_url(self, *args: Any) -> None:
        """
        Jump to log and search for next url
        """

        logger.debug("jumping to log and starting url search")
        self._go_log()
        self.conversation.wins.log_win.search_text = "http"
        self.conversation.wins.log_win.search_next()

    def _go_next(self, *args: Any) -> None:
        """
        Jump to a conversation with new messages or more recently used
        conversation
        """

        logger.debug("jumping to next conversation")

        # find next new conversation
        set_last_used = True
        conv = self.conversation.get_new()
        if not conv:
            # no new messages, try to go to next conversation
            logger.debug("conversation with new messages not found")
            conv = self.conversation.get_next()
            set_last_used = False
        if not conv:
            # nothing found, return
            logger.debug("next conversation not found")
            return

        # deactivate this and switch to other conversation
        self.go_back()
        conv.wins.list_win.jump_to_conv(conv, set_last_used=set_last_used)

    def _go_prev(self, *args: Any) -> None:
        """
        Jump to a previously used conversation
        """

        prev = self.conversation.get_prev()
        if not prev:
            return

        # deactivate this and switch to other conversation
        logger.debug("jumping to previous conversation")
        self.go_back()
        prev.wins.list_win.jump_to_conv(prev, set_last_used=False)

    def _go_conv(self, *args: Any) -> None:
        """
        Go to a specific conversation
        """

        # deactivate this and switch to filter mode in list window
        logger.debug("jumping to specific conversation")
        self.go_back()
        self.conversation.wins.list_win.go_conv()

    def _tab(self, *args: Any) -> None:
        # convert tab to spaces
        for _i in range(4):
            self.process_input(" ")

    def process_input(self, char: str) -> None:
        """
        Process user input (character)
        """

        segments = self.msg.split("\n")
        self.state.cur_y, self.state.cur_x = self.pad.getyx()
        pad_size_y, pad_size_x = self.pad.getmaxyx()

        # look for special key mappings in keymap or process as text
        if not self.handle_keybinds(char, segments):
            # insert new character into segments
            if not isinstance(char, str):
                return
            if char != "\n" and unicodedata.category(char)[0] == "C":
                return
            # make sure new char fits in the pad
            if len(segments) == pad_size_y - 1 and char == "\n":
                pad_size_y += 1
                self.pad.resize(pad_size_y, pad_size_x)
            if len(segments[self.state.cur_y]) == pad_size_x - 1 and \
               char != "\n":
                pad_size_x += 1
                self.pad.resize(pad_size_y, pad_size_x)

            segments[self.state.cur_y] = \
                segments[self.state.cur_y][:self.state.cur_x] + char +\
                segments[self.state.cur_y][self.state.cur_x:]
            # reconstruct orginal message for output in pad
            self.msg = "\n".join(segments)
            # reconstruct segments in case newline character was entered
            segments = self.msg.split("\n")
            # output new message in pad
            self.pad.erase()
            self.pad.addstr(self.msg)
            # move cursor to new position
            if char == "\n":
                self.pad.move(self.state.cur_y + 1, 0)
            else:
                self.pad.move(self.state.cur_y, self.state.cur_x + 1)
            # display changes in the pad
            self.redraw_pad()



"""
Nuqql UI Windows
"""

import curses
import logging
import re
import unicodedata

from typing import TYPE_CHECKING, Any, List

from .win import Win

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from nuqql.config import WinConfig  # noqa
    from nuqql.conversation import Conversation  # noqa

logger = logging.getLogger(__name__)


class ListWin(Win):
    """
    Class for List Windows
    """

    def __init__(self, config: "WinConfig", conversation: "Conversation",
                 title: str) -> None:
        Win.__init__(self, config, conversation, title)

        # filter for conversation list
        self.filter = ""
        self.filter_keyfunc = {
            "CURSOR_UP":    self._process_filter_up,
            "CURSOR_DOWN":  self._process_filter_down,
            "DEL_CHAR":     self._process_filter_del_char,
            "ENTER":        self._process_filter_enter,
            "GO_BACK":      self._process_filter_abort,
        }

        # list entries/message log
        self.list: List["Conversation"] = []

    def add(self, entry: "Conversation") -> None:
        """
        Add entry to internal list
        """

        # add entry to own list via base class function
        logger.debug("adding entry %s", entry)
        self.list_add(self.list, entry)

    def _match_filter(self, name: str) -> bool:
        """
        check if name matches the currently active filter
        """

        # no filter -> everything matches
        if not self.filter:
            return True

        # construct regurlar expression
        regex = ".*"
        for char in self.filter[1:].lower():
            regex += "{}.*".format(char)

        # check if any words in name match regular expression
        for word in name.lower().split():
            if re.match(regex, word):
                return True

        return False

    def _print_list(self, *args: Any) -> None:
        """
        Helper for printing the list
        """

        # parse arguments
        pad_size_x, pos_y, pos_x, win_size_y, win_size_x = args

        # is there a zoomed log window?
        zoomed_log_win = None

        # print names in list window
        for index, conv in enumerate(self.list):
            # get name of element; cut if it's too long
            name = conv.get_name()
            name = name[:pad_size_x-1] + "\n"

            # set colors depending on backend name
            assert conv.backend
            if conv.backend.name in self.config.attr["list_win_text"]:
                self.pad.attrset(self.config.attr["list_win_text"][
                    conv.backend.name])
            else:
                self.pad.attrset(self.config.attr["list_win_text"]["default"])

            # print name
            if index == self.state.cur_y:
                # cursor is on conversation, highlight it in list
                self.pad.insstr(index, 0, name, curses.A_REVERSE)
            else:
                # just show the conversation in list
                if self._match_filter(conv.get_name()):
                    self.pad.insstr(index, 0, name)
                else:
                    self.pad.insstr(index, 0, name, curses.A_DIM)

            # check if there is a zoomed conversation
            # TODO: move this into a separate helper?
            if conv.wins.log_win and conv.wins.log_win.state.zoomed:
                zoomed_log_win = conv.wins.log_win

        # move cursor back to original or active conversation's position
        self.pad.move(self.state.cur_y, self.state.cur_x)

        # check if visible part of pad needs to be moved and display it
        self._move_pad()
        self._check_borders()
        if zoomed_log_win:
            return
        self.pad.refresh(self.state.pad_y, self.state.pad_x,
                         pos_y + 1, pos_x + 1,
                         pos_y + win_size_y - 2,
                         pos_x + win_size_x - 2)

    def redraw_pad(self) -> None:
        """
        Redraw pad in window
        """

        # if terminal size is invalid, stop here
        if not self.config.is_terminal_valid():
            return

        # if list is empty, stop here
        if not self.list:
            return

        # screen/pad properties
        pos_y, pos_x = self.config.get_pos()
        win_size_y, win_size_x = self.win.getmaxyx()
        pad_size_y, pad_size_x = self.pad.getmaxyx()
        self.state.cur_y, self.state.cur_x = self.pad.getyx()

        # make sure pad has correct width (after resize)
        if pad_size_x != win_size_x - 2:
            self.pad.resize(pad_size_y, win_size_x - 2)
            pad_size_x = win_size_x - 2

        # store last selected entry
        if self.state.cur_y >= len(self.list):
            # make sure cur_y is still "within" the list. Length difference
            # should be only 1, because buddies get removed individually
            self.state.cur_y = max(0, self.state.cur_y - 1)
        last_selected = self.list[self.state.cur_y]

        # sort list
        self.list.sort()

        # make sure all names fit into pad
        if len(self.list) != pad_size_y:
            self.pad.resize(len(self.list), pad_size_x)

        # if there is an active conversation or last selected conversation was
        # moved, move cursor to it
        for index, conv in enumerate(self.list):
            if conv.is_active() or conv is last_selected:
                self.state.cur_y = index

        # print names in list window
        self._print_list(pad_size_x, pos_y, pos_x, win_size_y, win_size_x)

        # if in filter mode, show the filter string as well
        if self.filter:
            self._process_filter_show()

    def _cursor_top(self, *args: Any) -> None:
        # jump to first conversation
        if self.state.cur_y > 0:
            logger.debug("jumping to first conversation")
            self.pad.move(0, 0)

    def _cursor_bottom(self, *args: Any) -> None:
        # jump to last conversation
        lines = len(self.list)
        if self.state.cur_y < lines - 1:
            logger.debug("jumping to last conversation")
            self.pad.move(lines - 1, self.state.cur_x)

    def _cursor_page_up(self, *args: Any) -> None:
        # move cursor up one page until first entry in log
        win_size_y, unused_win_size_x = self.win.getmaxyx()

        if self.state.cur_y > 0:
            logger.debug("moving cursor one page up")
            if self.state.cur_y - (win_size_y - 2) >= 0:
                self.pad.move(self.state.cur_y - (win_size_y - 2),
                              self.state.cur_x)
            else:
                self.pad.move(0, self.state.cur_x)

    def _cursor_page_down(self, *args: Any) -> None:
        # move cursor down one page until last entry in log
        win_size_y, unused_win_size_x = self.win.getmaxyx()

        lines = len(self.list)
        if self.state.cur_y < lines:
            logger.debug("moving cursor one page down")
            if self.state.cur_y + win_size_y - 2 < lines:
                self.pad.move(self.state.cur_y + win_size_y - 2,
                              self.state.cur_x)
            else:
                self.pad.move(lines - 1, self.state.cur_x)

    def _cursor_up(self, *args: Any) -> None:
        # move cursor up until first entry in list
        if self.state.cur_y > 0:
            logger.debug("moving cursor up")
            self.pad.move(self.state.cur_y - 1, self.state.cur_x)

    def _cursor_down(self, *args: Any) -> None:
        # move cursor down until end of list
        if self.state.cur_y < len(self.list) - 1:
            logger.debug("moving cursor down")
            self.pad.move(self.state.cur_y + 1, self.state.cur_x)

    def _go_next(self, *args: Any) -> None:
        logger.debug("jumping to next conversation")

        # find a new(er) conversation and jump into it
        set_last_used = True
        conv = self.conversation.get_new()
        if not conv:
            logger.debug("conversation with new messages not found")
            conv = self.list[self.state.cur_y].get_next()
            set_last_used = False
        if not conv:
            logger.debug("next conversation not found")
            return
        self.jump_to_conv(conv, set_last_used=set_last_used)

    def _go_prev(self, *args: Any) -> None:
        # find older conversation and jump into it
        prev = self.list[self.state.cur_y].get_prev()
        if not prev:
            return

        # deactivate this and switch to other conversation
        logger.debug("jumping to previous conversation")
        prev.wins.list_win.jump_to_conv(prev, set_last_used=False)

    def _go_conv(self, *args: Any) -> None:
        # filter conversations and find specific conversation
        logger.debug("starting conversation filtering")
        self.filter = "/"

    def go_conv(self):
        """
        Helper for other windows for setting the window in filter mode
        """

        self._go_conv()
        self.redraw_pad()

    def _go_log(self, *args: Any) -> None:
        logger.debug("entering conversation's log")
        # go to conversation's log
        # create windows, if they do not exists
        if not self.list[self.state.cur_y].has_windows():
            self.list[self.state.cur_y].create_windows()
        # activate conversation's history
        self.list[self.state.cur_y].activate_log()

    def _enter(self, *args: Any) -> None:
        logger.debug("entering conversation")
        # enter conversation
        # create windows, if they do not exists
        if not self.list[self.state.cur_y].has_windows():
            self.list[self.state.cur_y].create_windows()
        # activate conversation
        self.list[self.state.cur_y].activate()
        # reset filter
        self.filter = ""

    def _quit(self, *args: Any) -> None:
        # quit nuqql
        logger.debug("quitting nuqql")
        self.state.active = False   # Exit the while loop

    def _process_filter_up(self) -> None:
        # move cursor up to next filter match
        conv_index = self.state.cur_y
        for index, conv in enumerate(self.list[:self.state.cur_y]):
            if self._match_filter(conv.get_name()):
                logger.debug("moving filter up to next match")
                conv_index = index

        self.pad.move(conv_index, self.state.cur_x)

    def _process_filter_down(self) -> None:
        # move cursor down to next filter match
        for index, conv in enumerate(self.list[self.state.cur_y + 1:]):
            if self._match_filter(conv.get_name()):
                logger.debug("moving filter down to next match")
                self.pad.move(index + self.state.cur_y + 1, self.state.cur_x)
                return

    def _process_filter_nearest(self) -> None:
        # move cursor do nearest filter match
        above = -1
        below = -1

        # find matches
        for index, conv in enumerate(self.list):
            if index <= self.state.cur_y and \
               self._match_filter(conv.get_name()):
                above = index
            if index > self.state.cur_y and \
               self._match_filter(conv.get_name()):
                below = index

        # is there a match above current cursor position?
        if above != -1:
            if below == -1:
                self.pad.move(above, self.state.cur_x)
            else:
                # matches above and below, what is nearer to current cursor?
                above_diff = above - self.state.cur_y
                below_diff = self.state.cur_y - below
                if above_diff < below_diff:
                    logger.debug("moving filter to nearest match above")
                    self.pad.move(above, self.state.cur_x)
                else:
                    logger.debug("moving filter to nearest match below")
                    self.pad.move(below, self.state.cur_x)
            return

        # nothing above current cursor position, below?
        if below != -1:
            logger.debug("moving filter to nearest match below")
            self.pad.move(below, self.state.cur_x)

    def _process_filter_show(self) -> None:
        """
        Show current filter string
        """

        if not self.filter:
            # show window border again
            self.redraw()
            return

        # TODO: use config get_size or something?
        max_y, max_x = self.win.getmaxyx()
        show = self.filter.ljust(max_x - 4, " ")
        self.win.addnstr(max_y - 1, 2, show, max_x - 4)
        self.win.refresh()

    def _process_filter_abort(self) -> None:
        # abort filter mode/reset filter
        logger.debug("leaving filter mode")
        self.filter = ""
        self._process_filter_show()

    def _process_filter_enter(self) -> None:
        # enter conversation in filter mode
        logger.debug("entering conversation from filter mode")

        # create windows, if they do not exists
        if not self.list[self.state.cur_y].has_windows():
            self.list[self.state.cur_y].create_windows()

        # activate conversation
        self.list[self.state.cur_y].activate()

        # reset filter
        self.filter = ""
        self._process_filter_show()

    def _process_filter_del_char(self) -> None:
        # delete character from filter
        logger.debug("deleting char from filter")
        self.filter = self.filter[:-1]
        self._process_filter_nearest()
        if not self.filter:
            # if filter is now empty, make sure it is not shown any more
            self._process_filter_show()

    def process_input(self, char: str) -> None:
        """
        Process input from user (character)
        """

        self.state.cur_y, self.state.cur_x = self.pad.getyx()

        # check if we are in filter mode
        if self.filter:
            # filter mode: look for special key mapping or process as text
            if not self.handle_keybinds(
                    char, keybinds=self.config.keybinds["__filter__"],
                    keyfunc=self.filter_keyfunc):
                # no special key, add character to filter
                try:
                    # filter special keys
                    if unicodedata.category(char)[0] == "C":
                        return
                    self.filter += char
                    self._process_filter_nearest()
                except (ValueError, TypeError):
                    pass
        else:
            # normal mode: look for special key mappings in keymap
            self.handle_keybinds(char)

        # display changes in the pad
        self.redraw_pad()

    def jump_to_conv(self, conversation: "Conversation",
                     set_last_used: bool = True) -> None:
        """
        Jump directly into specified conversation
        """

        logger.debug("jumping into conversation %s", conversation)

        # create conversation's windows if necessary
        if not conversation.has_windows():
            conversation.create_windows()

        # move cursor to this conversation
        for index, conv in enumerate(self.list):
            if conv is conversation:
                self.state.cur_y = index
                self.pad.move(self.state.cur_y, self.state.cur_x)

        # finally, activate conversation
        conversation.activate(set_last_used=set_last_used)



"""
Dummy Nuqql Backend
"""

import logging

from pathlib import Path
from typing import Callable, List

from .backend import Backend

logger = logging.getLogger(__name__)


class NuqqlBackend(Backend):
    """
    Class for the nuqql dummy backend
    """

    def __init__(self, name: str) -> None:
        Backend.__init__(self, name)
        self.version = ""

        # function for (re)starting backends
        self.restart_func: Callable[[str], None]

    def _handle_nuqql_global_status(self, parts: List[str]) -> None:
        """
        Handle nuqql command: global-status
        Call getter and setter funcions
        """

        if not parts:
            return
        sub_command = parts[0]
        if sub_command == "set":
            if len(parts) < 2:
                return
            self._handle_nuqql_global_status_set(parts[1])
        elif sub_command == "get":
            self._handle_nuqql_global_status_get()

    def _handle_nuqql_global_status_set(self, status: str) -> None:
        """
        Handle nuqql command: global-status set
        Set status and store it in global_status file
        """

        # only use the first word as status
        if not status or status == "":
            return

        logger.debug("setting global status to %s", status)

        # write status
        self._write_global_status(status)

        # set status in all backends and their accounts
        for backend in self.backends.values():
            for acc in backend.accounts.values():
                if backend.client:
                    backend.client.send_status_set(acc.aid, status)

        # log message
        msg = "global-status: " + status
        if self.conversation:
            self.conversation.log("nuqql", msg)

    def _handle_nuqql_global_status_get(self) -> None:
        """
        Handle nuqql command: global-status get
        Read status from global_status file
        """

        logger.debug("getting global status")

        # read status
        status = self.read_global_status()
        if status == "":
            return
        logger.debug("global status is %s", status)

        # log message
        msg = "global-status: " + status
        if self.conversation:
            self.conversation.log("nuqql", msg)

    @staticmethod
    def _write_global_status(status: str) -> None:
        """
        Write global status to global_status file
        """

        logger.debug("writing global status %s to file", status)

        # write status to file
        global_status_dir = str(Path.home()) + "/.config/nuqql"
        Path(global_status_dir).mkdir(parents=True, exist_ok=True)
        global_status_file = global_status_dir + "/global_status"
        line = status + "\n"
        lines = []
        lines.append(line)
        with open(global_status_file, "w+") as status_file:
            status_file.writelines(lines)

    def _handle_stop(self, parts: List[str]) -> None:
        """
        Handle stop command, stop a backend
        """

        if not parts:
            return

        backend_name = parts[0]
        logger.debug("stopping backend %s", backend_name)
        if backend_name in self.backends:
            self.backends[backend_name].stop()

    def _handle_start(self, parts: List[str]) -> None:
        """
        Handle start command, start a backend
        """

        if not parts:
            return

        backend_name = parts[0]
        logger.debug("starting backend %s", backend_name)
        assert self.restart_func
        self.restart_func(backend_name)

    def _handle_restart(self, parts: List[str]) -> None:
        """
        Handle restart command, stop and start a backend
        """

        if not parts:
            return

        backend_name = parts[0]
        logger.debug("restarting backend %s", backend_name)
        self._handle_stop(parts)
        self._handle_start(parts)

    def _handle_quit(self, _parts: List[str]) -> None:
        """
        Handle quit command, quit nuqql
        """

        logger.debug("quitting nuqql")
        if self.conversation:
            self.conversation.wins.input_win.state.active = False
            self.conversation.wins.list_win.state.active = False

    def _handle_version(self, _parts: List[str]) -> None:
        """
        Handle version command, print nuqql version
        """

        # log message
        msg = f"version: nuqql v{self.version}"
        logger.debug("getting nuqql version: %s", msg)
        if self.conversation:
            self.conversation.log("nuqql", msg)

    def handle_nuqql_command(self, msg: str) -> None:
        """
        Handle a nuqql command (from the nuqql conversation)
        """

        logger.debug("handling nuqql command %s", msg)

        # parse message
        parts = msg.split()
        if not parts:
            return

        # check command and call helper functions
        command_map = {
            "global-status": self._handle_nuqql_global_status,
            "stop": self._handle_stop,
            "start": self._handle_start,
            "restart": self._handle_restart,
            "quit": self._handle_quit,
            "version": self._handle_version,
        }
        command = parts[0]
        if command in command_map:
            command_map[command](parts[1:])



"""
nuqql configurations
"""

import configparser
import logging

from pathlib import Path
from typing import Dict, Any

logger = logging.getLogger(__name__)

# configurations
CONFIGS: Dict[str, Any] = {}


def read_from_file() -> configparser.ConfigParser:
    """
    Read configuration from config file
    """
    config_file = Path.home() / ".config/nuqql/config.ini"
    config = configparser.ConfigParser()
    config.optionxform = lambda option: option  # type: ignore
    config.read(config_file)

    return config


def write_to_file(config: configparser.ConfigParser) -> None:
    """
    Write configuration to config file
    """

    config_file = Path.home() / ".config/nuqql/config.ini"
    with open(config_file, "w+") as configfile:
        config.write(configfile)


def get(name: str) -> Any:
    """
    Get configuration identified by name
    """

    return CONFIGS[name]



"""
Log message
"""

import datetime


class LogMessage:
    """Class for log messages to be displayed in LogWins"""

    def __init__(self, tstamp: datetime.datetime, sender: str, msg: str,
                 own: bool = False) -> None:
        """
        Initialize log message with timestamp, sender of the message, and
        the message itself
        """

        # timestamp
        self.tstamp = tstamp

        # sender could be us or buddy/other user, as
        # indicated by self.own (helps with coloring etc. later)
        self.sender = sender
        self.own = own

        # the message itself
        self.msg = msg

        # has message been read?
        self.is_read = False

    def get_short_sender(self) -> str:
        """
        Convert name to a shorter version
        """

        # TODO: improve? Save short name in account and buddy instead?
        return self.sender.split("@")[0]

    def read(self, mark_read: bool = True) -> str:
        """
        Format and return log message; mark it as read
        """

        # format message
        msg = "{0} {1}: {2}".format(self.tstamp.strftime("%H:%M:%S"),
                                    self.get_short_sender(),
                                    self.msg)

        # message has now been read
        if mark_read:
            self.is_read = True

        return msg

    def is_equal(self, other: "LogMessage") -> bool:
        """
        Check if this message and the LogMessage "other" match
        """

        if self.tstamp != other.tstamp:
            return False
        if self.sender != other.sender:
            return False
        if self.msg != other.msg:
            return False

        return True



"""
Main part of nuqql.
"""

import logging
import os
import signal

import nuqql.backend
import nuqql.config
import nuqql.ui

logger = logging.getLogger(__name__)


# main loop of nuqql
def main_loop() -> str:
    """
    Main loop of nuqql.
    """

    logger.debug("entering main loop")
    try:
        # init and start all backends
        nuqql.backend.start_backends()

        # loop as long as user does not quit
        while nuqql.ui.handle_input():
            # update buddies
            nuqql.backend.update_buddies()

            # handle network input
            nuqql.backend.handle_network()
    finally:
        # shut down backends
        nuqql.backend.stop_backends()

    # quit nuqql
    return ""


def _set_esc_delay() -> None:
    """
    Configure ESC delay for curses
    """

    os.environ.setdefault("ESCDELAY", nuqql.config.get("escdelay"))


# main entry point
def run() -> None:
    """
    Main entry point of nuqql
    """

    # does not go to nuqql log file
    logger.debug("starting nuqql")

    # parse command line arguments
    nuqql.config.parse_args()

    # ignore SIGINT
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    # configure esc delay for curses
    _set_esc_delay()

    # initialize ui and run main_loop
    nuqql.ui.init(main_loop)



"""
nuqql window config
"""

import curses
import curses.ascii
import logging
import sys

from typing import Any, Dict, Tuple

from .configs import get, read_from_file, write_to_file

logger = logging.getLogger(__name__)

# default keymap for special keys
DEFAULT_KEYMAP: Dict[str, Any] = {
    "KEY_ENTER":        ord("\n"),
    "KEY_ESC":          curses.ascii.ESC,
    "KEY_RIGHT":        curses.KEY_RIGHT,
    "KEY_LEFT":         curses.KEY_LEFT,
    "KEY_DOWN":         curses.KEY_DOWN,
    "KEY_UP":           curses.KEY_UP,
    "KEY_H":            ord("h"),
    "KEY_J":            ord("j"),
    "KEY_K":            ord("k"),
    "KEY_L":            ord("l"),
    "KEY_N":            ord("n"),
    "KEY_P":            ord("p"),
    "KEY_Q":            ord("q"),
    "KEY_TAB":          ord("\t"),
    "KEY_CTRL_A":       ord(curses.ascii.ctrl("a")),
    "KEY_CTRL_B":       ord(curses.ascii.ctrl("b")),
    "KEY_CTRL_E":       ord(curses.ascii.ctrl("e")),
    "KEY_CTRL_K":       ord(curses.ascii.ctrl("k")),
    "KEY_CTRL_N":       ord(curses.ascii.ctrl("n")),
    "KEY_CTRL_O":       ord(curses.ascii.ctrl("o")),
    "KEY_CTRL_U":       ord(curses.ascii.ctrl("u")),
    "KEY_CTRL_V":       ord(curses.ascii.ctrl("v")),
    "KEY_CTRL_X":       ord(curses.ascii.ctrl("x")),
    "KEY_DEL":          curses.ascii.DEL,
    "KEY_BS":           curses.ascii.BS,
    "KEY_BACKSPACE":    curses.KEY_BACKSPACE,
    "KEY_DC":           curses.KEY_DC,
    "KEY_HOME":         curses.KEY_HOME,
    "KEY_END":          curses.KEY_END,
    "KEY_PAGE_UP":      curses.KEY_PPAGE,
    "KEY_PAGE_DOWN":    curses.KEY_NPAGE,
    "KEY_F9":           curses.KEY_F9,
    "KEY_F10":          curses.KEY_F10,
    "KEY_/":            ord("/"),
}

# default key bindings for input windows
DEFAULT_INPUT_WIN_KEYBINDS = {
    "CURSOR_RIGHT":         "KEY_RIGHT",
    "CURSOR_LEFT":          "KEY_LEFT",
    "CURSOR_DOWN":          "KEY_DOWN",
    "CURSOR_UP":            "KEY_UP",
    "CURSOR_LINE_START":    "KEY_PAGE_UP",
    "CURSOR_LINE_END":      "KEY_PAGE_DOWN",
    "CURSOR_MSG_START":     "KEY_HOME, KEY_CTRL_A",
    "CURSOR_MSG_END":       "KEY_END, KEY_CTRL_E",
    "DEL_CHAR":             "KEY_DEL, KEY_BACKSPACE, KEY_BS",
    "DEL_CHAR_RIGHT":       "KEY_DC",
    "DEL_LINE_END":         "KEY_CTRL_K",
    "DEL_LINE":             "KEY_CTRL_U",
    "GO_BACK":              "KEY_ESC",
    "GO_NEXT":              "KEY_CTRL_N",
    "GO_PREV":              "KEY_CTRL_B",
    "GO_CONV":              "KEY_CTRL_V",
    "GO_LOG":               "KEY_CTRL_O, KEY_F9",
    "GO_LOG_SEARCH_URL":    "KEY_F10",
    "SEND_MSG":             "KEY_CTRL_X",
    "TAB":                  "KEY_TAB",
}

# default key bindings for log windows
DEFAULT_LOG_WIN_KEYBINDS = {
    "CURSOR_RIGHT":         "KEY_RIGHT, KEY_L",
    "CURSOR_LEFT":          "KEY_LEFT, KEY_H",
    "CURSOR_DOWN":          "KEY_DOWN, KEY_J",
    "CURSOR_UP":            "KEY_UP, KEY_K",
    "CURSOR_TOP":           "KEY_HOME, KEY_CTRL_A",
    "CURSOR_BOTTOM":        "KEY_END, KEY_CTRL_E",
    "CURSOR_PAGE_UP":       "KEY_PAGE_UP",
    "CURSOR_PAGE_DOWN":     "KEY_PAGE_DOWN",
    "GO_BACK":              "KEY_ESC, KEY_F9",
    "SEARCH":               "KEY_/",
    "SEARCH_NEXT":          "KEY_N, KEY_F10",
    "SEARCH_PREV":          "KEY_P",
    "SEND_MSG":             "KEY_CTRL_X",
}

# default key bindings for log window in search input mode
DEFAULT_LOG_WIN_SEARCH_KEYBINDS = {
    "GO_BACK":              "KEY_ESC",
    "ENTER":                "KEY_ENTER",
    "DEL_CHAR":             "KEY_DEL, KEY_BACKSPACE, KEY_BS",
}

# default key bindings for list window (Buddy List)
DEFAULT_LIST_WIN_KEYBINDS = {
    "CURSOR_DOWN":          "KEY_DOWN, KEY_J",
    "CURSOR_UP":            "KEY_UP, KEY_K",
    "CURSOR_TOP":           "KEY_HOME, KEY_CTRL_A",
    "CURSOR_BOTTOM":        "KEY_END, KEY_CTRL_E",
    "CURSOR_PAGE_UP":       "KEY_PAGE_UP",
    "CURSOR_PAGE_DOWN":     "KEY_PAGE_DOWN",
    "ENTER":                "KEY_ENTER",
    "GO_BACK":              "KEY_ESC",
    "GO_NEXT":              "KEY_CTRL_N",
    "GO_PREV":              "KEY_CTRL_B",
    "GO_CONV":              "KEY_CTRL_V, KEY_/",
    "GO_LOG":               "KEY_H",
    "QUIT":                 "KEY_Q",
}

# default key bindings for list window in filter mode
DEFAULT_LIST_WIN_FILTER_KEYBINDS = {
    "CURSOR_DOWN":          "KEY_DOWN",
    "CURSOR_UP":            "KEY_UP",
    "GO_BACK":              "KEY_ESC",
    "ENTER":                "KEY_ENTER",
    "DEL_CHAR":             "KEY_DEL, KEY_BACKSPACE, KEY_BS",
}

# default ui layout
DEFAULT_LAYOUT: Dict[str, Any] = {
    # window x and y sizes in percent
    "LIST_WIN_Y_PER":       100,
    "LIST_WIN_X_PER":       20,
    "LOG_WIN_Y_PER":        80,
    "LOG_WIN_X_PER":        80,
    "LOG_WIN_MAIN_Y_PER":   100,
    "LOG_WIN_MAIN_X_PER":   80,
    "INPUT_WIN_Y_PER":      20,
    "INPUT_WIN_X_PER":      80,
}

# window default colors and attributes
DEFAULT_COLORS = {
    "background": "default",
    "win_border": "blue",
    "list_win_text_default": "yellow",
    "list_win_text_nuqql": "yellow",
    "list_win_text_slixmppd": "green",
    "list_win_text_matrixd": "red",
    "list_win_text_matrixd-nio": "red",
    "list_win_text_mattermostd": "cyan",
    "list_win_text_purpled": "magenta",
    "list_win_text_based": "white",
    "log_win_text_peer_old": "yellow",
    "log_win_text_peer_new": "yellow",
    "log_win_text_self_old": "cyan",
    "log_win_text_self_new": "cyan",
}

DEFAULT_ATTRIBS = {
    "win_border": "bold",
    "list_win_text_default": "normal",
    "list_win_text_nuqql": "normal",
    "list_win_text_slixmppd": "normal",
    "list_win_text_matrixd": "normal",
    "list_win_text_matrixd-nio": "normal",
    "list_win_text_mattermostd": "normal",
    "list_win_text_purpled": "normal",
    "list_win_text_based": "normal",
    "log_win_text_peer_old": "normal",
    "log_win_text_peer_new": "bold",
    "log_win_text_self_old": "normal",
    "log_win_text_self_new": "bold",
}
# default window settings
DEFAULT_LIST_WIN_CONFIG: Dict[str, Any] = {
    "show_title": True,
}

DEFAULT_LOG_WIN_CONFIG = DEFAULT_LIST_WIN_CONFIG
DEFAULT_INPUT_WIN_CONFIG = DEFAULT_LIST_WIN_CONFIG


class WinConfig:
    """
    Class for window configuration
    """

    def __init__(self, wtype: str) -> None:
        self._type = wtype
        self._rel_yx = 0.0, 0.0
        self.keymap: Dict[int, Any] = {}
        self.keybinds: Dict[str, Any] = {}
        self.attr: Dict[str, Any] = {}  # window colors/attributes
        self.settings: Dict[str, str] = {}  # window settings

    @staticmethod
    def _get_layout_config() -> Dict:
        """
        Initialize/get layout configuration
        """

        # init configuration from defaults
        layout_config = DEFAULT_LAYOUT

        # parse config read from file
        config = read_from_file()
        for section in config.sections():
            if section == "layout":
                # overwrite default color config entries
                for key in config["layout"]:
                    if key in layout_config:
                        layout_config[key] = float(config["layout"][key])

        # write (updated) config to file again
        config["layout"] = layout_config
        write_to_file(config)

        # create and return internally used layout configuration
        layout = {}
        for key, value in layout_config.items():
            layout[key] = value / 100
        return layout

    def init_layout(self) -> None:
        """
        Init UI/Window layout
        """

        # get layout
        layout = self._get_layout_config()

        # set window layout
        if self._type == "list_win":
            self._rel_yx = layout["LIST_WIN_Y_PER"], layout["LIST_WIN_X_PER"]
        if self._type == "log_win":
            self._rel_yx = layout["LOG_WIN_Y_PER"], layout["LOG_WIN_X_PER"]
        if self._type == "log_win_main":
            self._rel_yx = layout["LOG_WIN_MAIN_Y_PER"], \
                    layout["LOG_WIN_MAIN_X_PER"]
        if self._type == "input_win":
            self._rel_yx = layout["INPUT_WIN_Y_PER"], layout["INPUT_WIN_X_PER"]

        logger.debug("initialized window layout of %s", self._type)

    @staticmethod
    def _get_window_config() -> Dict[str, Any]:
        """
        Initialize/get window configuration
        """

        # init configuration from defaults
        win_config = {}
        win_config["list_win"] = DEFAULT_LIST_WIN_CONFIG
        win_config["log_win"] = DEFAULT_LOG_WIN_CONFIG
        win_config["log_win_main"] = DEFAULT_LOG_WIN_CONFIG
        win_config["input_win"] = DEFAULT_INPUT_WIN_CONFIG

        # parse config read from file
        config = read_from_file()
        for section in config.sections():
            if section in ("list_win", "log_win", "log_win_main", "input_win"):
                # overwrite default config entries
                for key in config[section]:
                    if key in win_config[section]:
                        try:
                            # try to read setting as bool
                            win_config[section][key] = \
                                    config[section].getboolean(key)
                        except ValueError:
                            win_config[section][key] = config[section][key]

        # write (updated) config to file again
        for win_type in ("list_win", "log_win", "log_win_main", "input_win"):
            config[win_type] = win_config[win_type]
        write_to_file(config)

        return win_config

    def init_window_settings(self) -> None:
        """
        Init window specific settings
        """

        # get window configurations
        win_config = self._get_window_config()

        # use settings depending on window type
        self.settings = win_config[self._type]

        logger.debug("initialized window specific settings of %s", self._type)

    @staticmethod
    def _get_color_config() -> Tuple[Dict[str, str], Dict[str, str]]:
        """
        Initialize/get color configuration
        """

        # init configuration from defaults
        color_config = DEFAULT_COLORS
        attrib_config = DEFAULT_ATTRIBS

        # parse config read from file
        config = read_from_file()
        for section in config.sections():
            if section == "colors":
                # overwrite default color config entries
                for key in config["colors"]:
                    if key in color_config:
                        color_config[key] = config["colors"][key]
            if section == "attributes":
                # overwrite default attribute config entries
                for key in config["attributes"]:
                    if key in attrib_config:
                        attrib_config[key] = config["attributes"][key]

        # write (updated) config to file again
        config["colors"] = color_config
        config["attributes"] = attrib_config
        write_to_file(config)

        return color_config, attrib_config

    def init_colors(self) -> None:
        """
        Initialize colors
        """

        # get color and attrib configuration
        color_config, attrib_config = self._get_color_config()

        # configure background color
        background = color_config["background"]
        bg_colors = {
            "default":  int(-1),
            "black":    curses.COLOR_BLACK,
            "blue":     curses.COLOR_BLUE,
            "cyan":     curses.COLOR_CYAN,
            "green":    curses.COLOR_GREEN,
            "magenta":  curses.COLOR_MAGENTA,
            "red":      curses.COLOR_RED,
            "white":    curses.COLOR_WHITE,
            "yellow":   curses.COLOR_YELLOW,
        }
        # allow definition of color numbers
        for i in range(0, 16):
            bg_colors["color{}".format(i)] = i

        # allow usage of default colors and initialize color pairs
        curses.use_default_colors()
        curses.init_pair(1, curses.COLOR_BLACK, bg_colors[background])
        curses.init_pair(2, curses.COLOR_BLUE, bg_colors[background])
        curses.init_pair(3, curses.COLOR_CYAN, bg_colors[background])
        curses.init_pair(4, curses.COLOR_GREEN, bg_colors[background])
        curses.init_pair(5, curses.COLOR_MAGENTA, bg_colors[background])
        curses.init_pair(6, curses.COLOR_RED, bg_colors[background])
        curses.init_pair(7, curses.COLOR_WHITE, bg_colors[background])
        curses.init_pair(8, curses.COLOR_YELLOW, bg_colors[background])
        # allow definition of additional (unnamed) colors
        for i in range(8, 16):
            curses.init_pair(i + 1, i, bg_colors[background])

        # mapping from color names to color pairs
        colors = {
            "default":  curses.color_pair(0),
            "black":    curses.color_pair(1),
            "blue":     curses.color_pair(2),
            "cyan":     curses.color_pair(3),
            "green":    curses.color_pair(4),
            "magenta":  curses.color_pair(5),
            "red":      curses.color_pair(6),
            "white":    curses.color_pair(7),
            "yellow":   curses.color_pair(8),
        }
        # allow definition of color numbers
        for i in range(0, 16):
            colors["color{}".format(i)] = curses.color_pair(i + 1)

        # text attributes
        attrib_italic = curses.A_NORMAL     # italic was added in python 3.7
        if sys.version_info >= (3, 7):
            attrib_italic = curses.A_ITALIC
        attribs = {
            "alt":          curses.A_ALTCHARSET,
            "blink":        curses.A_BLINK,
            "bold":         curses.A_BOLD,
            "dim":          curses.A_DIM,
            "invisible":    curses.A_INVIS,
            "italic":       attrib_italic,  # italic was added in python 3.7
            "normal":       curses.A_NORMAL,
            "protect":      curses.A_PROTECT,
            "reverse":      curses.A_REVERSE,
            "standout":     curses.A_STANDOUT,
            "underline":    curses.A_UNDERLINE,
            "horizontal":   curses.A_HORIZONTAL,
            "left":         curses.A_LEFT,
            "low":          curses.A_LOW,
            "right":        curses.A_RIGHT,
            "top":          curses.A_TOP,
            "vertical":     curses.A_VERTICAL,
        }

        # actually create color/attrib config for later use
        self.attr = {}
        self.attr["list_win_text"] = {}

        for key, value in color_config.items():
            # read attribute if it exists (e.g., background only has a color)
            attrib = 0
            if key in attrib_config:
                attrib = attribs[attrib_config[key]]

            # list_win_text needs special handling
            if key.startswith("list_win_text_"):
                self.attr["list_win_text"][key[14:]] = \
                    colors[value] | attrib
                # skip to next key
                continue

            # set color and attrib in configuration
            self.attr[key] = colors[value] | attrib

        logger.debug("initialized colors of %s", self._type)

    @staticmethod
    def _get_keymap_config() -> Dict[int, Any]:
        """
        Initialize/get keymap configuration
        """

        # init configuration from defaults
        keymap_config = DEFAULT_KEYMAP

        # parse config read from file
        config = read_from_file()
        for section in config.sections():
            if section == "keymap":
                # overwrite default keymap config entries
                for key in config["keymap"]:
                    if key in keymap_config:
                        try:
                            keymap_config[key] = int(config["keymap"][key])
                        except ValueError:
                            keymap_config[key] = config["keymap"][key]

        # write (updated) config to file again
        config["keymap"] = keymap_config
        write_to_file(config)

        # create and return internally used keymap
        keymap = {}
        for key, value in keymap_config.items():
            keymap[value] = key
        return keymap

    def init_keymap(self) -> None:
        """
        Initialzize keymap
        """

        self.keymap = self._get_keymap_config()
        logger.debug("initialized keymap of %s", self._type)

    @staticmethod
    def _get_keybinds_config() -> Dict[str, Any]:
        """
        Initialize/get keybind configuration
        """

        # init configuration from defaults
        keybinds_config: Dict[str, Any] = {}
        keybinds_config["list_win_keybinds"] = DEFAULT_LIST_WIN_KEYBINDS
        keybinds_config["list_win_filter_keybinds"] = \
            DEFAULT_LIST_WIN_FILTER_KEYBINDS
        keybinds_config["log_win_keybinds"] = DEFAULT_LOG_WIN_KEYBINDS
        keybinds_config["log_win_search_keybinds"] = \
            DEFAULT_LOG_WIN_SEARCH_KEYBINDS
        keybinds_config["input_win_keybinds"] = DEFAULT_INPUT_WIN_KEYBINDS

        # parse config read from file
        config = read_from_file()
        for section in config.sections():
            if section in ("list_win_keybinds", "list_win_filter_keybinds",
                           "log_win_keybinds", "log_win_search_keybinds",
                           "input_win_keybinds"):
                # overwrite default keymap config entries
                for key in config[section]:
                    if key in keybinds_config[section]:
                        keybinds_config[key] = config[section][key]

        # write (updated) config to file again
        for win_binds in ("list_win_keybinds", "list_win_filter_keybinds",
                          "log_win_keybinds", "log_win_search_keybinds",
                          "input_win_keybinds"):
            config[win_binds] = keybinds_config[win_binds]
        write_to_file(config)

        # create and return internally used keymap
        # swap keys and values in config dictionaries
        tmp_keybinds: Dict[str, Any] = {}
        for win_binds in ("list_win_keybinds", "list_win_filter_keybinds",
                          "log_win_keybinds", "log_win_search_keybinds",
                          "input_win_keybinds"):
            tmp_keybinds[win_binds] = {}
            for key, value in keybinds_config[win_binds].items():
                # might be comma separated list of keys bound to same function
                for entry in value.split(","):
                    tmp_keybinds[win_binds][entry.strip()] = key

        # construct keybinds as used later
        keybinds = {}
        keybinds["list_win"] = tmp_keybinds["list_win_keybinds"]
        # list win has additional keybinds for filter mode
        keybinds["list_win"]["__filter__"] = \
            tmp_keybinds["list_win_filter_keybinds"]
        keybinds["log_win"] = tmp_keybinds["log_win_keybinds"]
        keybinds["log_win_main"] = tmp_keybinds["log_win_keybinds"]
        # log win has additional keybinds for search input mode
        keybinds["log_win"]["__search__"] = \
            tmp_keybinds["log_win_search_keybinds"]
        keybinds["log_win_main"]["__search__"] = \
            tmp_keybinds["log_win_search_keybinds"]
        keybinds["input_win"] = tmp_keybinds["input_win_keybinds"]

        return keybinds

    def init_keybinds(self) -> None:
        """
        Initialize keybindings depending on window type
        """

        keybinds = self._get_keybinds_config()
        self.keybinds = keybinds[self._type]
        logger.debug("initialized key binds of %s", self._type)

    def get_win_size(self, max_y: int, max_x: int) -> Tuple[int, int]:
        """
        Get size of the window, depending on max_y and max_x as well as
        window's y_per and x_per. If window size is smaller than minimum size,
        return minimum size.
        """

        abs_y = max(int(max_y * self._rel_yx[0]), 3)
        abs_x = max(int(max_x * self._rel_yx[1]), 3)
        return abs_y, abs_x

    def get_size(self) -> Tuple[int, int]:
        """
        Return window size depending on max screen size and
        other windows' sizes.
        """

        # get (minimum) size of all windows
        max_y, max_x = get("screen").getmaxyx()
        list_y, list_x = get("list_win").get_win_size(max_y, max_x)
        log_y, log_x = get("log_win").get_win_size(max_y, max_x)
        log_main_y, log_main_x = get("log_win_main").get_win_size(max_y, max_x)
        input_y, input_x = get("input_win").get_win_size(max_y, max_x)

        # make sure input and log win use all available space
        if input_y + log_y < max_y:
            log_y += max_y - input_y

        # reduce log window height if necessary
        if input_y + log_y > max_y:
            log_y = max(max_y - input_y, 3)

        # reduce log main window height if necessary
        if log_main_y > max_y:
            log_main_y = max(max_y, 3)

        # reduce log and input window width if necessary
        if list_x + log_x > max_x:
            log_x = max(max_x - list_x, 3)
            input_x = max(max_x - list_x, 3)

        # reduce log main window width if necessary
        if list_x + log_main_x > max_x:
            log_main_x = max(max_x - list_x, 3)

        # return height and width of this window
        if self._type == "log_win":
            return log_y, log_x
        if self._type == "log_win_main":
            return log_main_y, log_main_x
        if self._type == "input_win":
            return input_y, input_x
        if self._type == "list_win":
            return list_y, list_x

        # should not be reached
        return -1, -1

    def get_pos(self) -> Tuple[int, int]:
        """
        Get position of the window, depending on type and window sizes
        """

        max_y, max_x = get("screen").getmaxyx()
        size_y, size_x = self.get_size()

        if self._type == "list_win":
            return 0, 0
        if self._type == "log_win":
            return 0, max_x - size_x
        if self._type == "log_win_main":
            return 0, max_x - size_x
        if self._type == "input_win":
            return max_y - size_y, max_x - size_x

        # should not be reached
        return -1, -1

    @staticmethod
    def is_terminal_valid() -> bool:
        """
        Helper that checks if terminal size is still valid (after resize)
        """

        # height and width of screen should not get below minimum size
        max_y, max_x = get("screen").getmaxyx()
        if max_y < 6 or max_x < 6:
            logger.debug("terminal is invalid")
            return False

        # everything seems to be ok
        return True



"""
Backend client
"""

import logging
import socket
import select
import time
import html

from typing import Optional, TYPE_CHECKING

import nuqql.conversation

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from nuqql.backend import Backend

logger = logging.getLogger(__name__)

# how often should a backend client try to connect to its server and
# how long (in seconds) should a backend client sleep between retries?
CLIENT_MAX_RETRIES = 100
CLIENT_RETRY_SLEEP = 0.1

# backend error message
BACKEND_ERROR = "Error accessing backend."

# network buffer
BUFFER_SIZE = 4096


class BackendClient:
    """
    Class for a backend's client connection to a
    local or remote backend server process
    """

    def __init__(self, sock_af: "socket.AddressFamily" = socket.AF_UNIX,
                 ip_addr: str = "127.0.0.1", port: int = 32000,
                 sock_file: str = "") -> None:
        # client
        self.backend: Optional["Backend"] = None
        self.sock: Optional[socket.socket] = None
        self.sock_af = sock_af
        self.sock_file = sock_file
        self.ip_addr = ip_addr
        self.port = port
        self.buffer = ""

    def _connect(self) -> None:
        """
        Helper for connecting to the server
        """

        if self.sock_af == socket.AF_INET:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.connect((self.ip_addr, self.port))
            logger.debug("connected AF_INET socket")
        elif self.sock_af == socket.AF_UNIX:
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.sock.connect(self.sock_file)
            logger.debug("connected AF_UNIX socket")

    def start(self) -> None:
        """
        Start the backend's client
        """

        # open sockets and connect
        retries = 0
        logger.debug("starting client")
        while not self.sock and retries < CLIENT_MAX_RETRIES:
            try:
                self._connect()
            except OSError:
                self.sock = None
                retries += 1
                time.sleep(CLIENT_RETRY_SLEEP)

    def stop(self) -> None:
        """
        Stop the backend's client
        """

        logger.debug("stopping client")
        if self.sock:
            try:
                self.sock.close()
            except OSError:
                pass
            self.sock = None

    def read(self) -> Optional[str]:
        """
        Read from the client connection
        """

        if not self.sock:
            return None

        try:
            reads, unused_writes, errs = select.select([self.sock, ], [],
                                                       [self.sock, ], 0)
        except OSError:
            nuqql.conversation.log_nuqql_conv(BACKEND_ERROR)
            logger.error("read error (select)")
            if self.backend:
                self.backend.stop()
            return None

        if self.sock in errs:
            # something is wrong
            logger.error("read error (socket)")
            if self.backend:
                self.backend.stop()
            return None

        if self.sock in reads:
            # read data from socket and add it to buffer
            try:
                data = self.sock.recv(BUFFER_SIZE)
            except OSError:
                nuqql.conversation.log_nuqql_conv(BACKEND_ERROR)
                logger.error("read error (recv)")
                if self.backend:
                    self.backend.stop()
                return None
            self.buffer += data.decode()

        # get next message from buffer and return it
        eom = self.buffer.find("\r\n")
        if eom == -1:
            # no message found
            return None

        # remove message from buffer and return it
        msg = self.buffer[:eom]
        # remove message including "\r\n" from buffer
        self.buffer = self.buffer[eom + 2:]

        logger.debug("read message: %s", msg)
        return msg

    def _send(self, msg: str) -> None:
        """
        Helper for sending any messages and catching errors.
        """

        if not self.sock:
            return

        try:
            self.sock.sendall(msg.encode())
            logger.debug("sent message: %s", msg)
        except OSError:
            nuqql.conversation.log_nuqql_conv(BACKEND_ERROR)
            logger.error("send error")
            if self.backend:
                self.backend.stop()
            return

    def send_command(self, cmd: str) -> None:
        """
        Send a command over the client connection
        """

        msg = cmd + "\r\n"
        logger.debug("sending command: %s", msg)
        self._send(msg)

    def send_msg(self, account_id: str, buddy: str, msg: str) -> None:
        """
        Send a regular message over the client connection
        """

        prefix = "account {0} send {1} ".format(account_id, buddy)
        msg = html.escape(msg)
        msg = "<br/>".join(msg.split("\n"))
        msg = prefix + msg + "\r\n"
        logger.debug("sending message: %s", msg)
        self._send(msg)

    def send_group_msg(self, account_id: str, buddy: str, msg: str) -> None:
        """
        Send a group message over the client connection
        """

        prefix = "account {0} chat send {1} ".format(account_id, buddy)
        msg = html.escape(msg)
        msg = "<br/>".join(msg.split("\n"))
        msg = prefix + msg + "\r\n"
        logger.debug("sending group message: %s", msg)
        self._send(msg)

    def send_collect(self, account_id: str) -> None:
        """
        Send "collect" message over the client connection,
        which collects all messages received by the backend
        """

        # collect all messages since time 0
        # TODO: only works as intended if we spawn our own purpled daemon at
        # nuqql's startup, FIXME?
        msg = "account {0} collect 0\r\n".format(account_id)
        # self.collect_acc = account
        logger.debug("sending collect message: %s", msg)
        self._send(msg)

    def send_buddies(self, account_id: str) -> None:
        """
        Send "buddies" message over the client connection,
        which retrieves all buddies of the specified account from the backend
        """

        msg = "account {0} buddies\r\n".format(account_id)
        logger.debug("sending buddies message: %s", msg)
        self._send(msg)

    def send_accounts(self) -> None:
        """
        Send "account list" message over the client connection,
        which retrieves all accounts from the backend
        """

        msg = "account list\r\n"
        logger.debug("sending account list message: %s", msg)
        self._send(msg)

    def send_status_set(self, account_id: str, status: str) -> None:
        """
        Send "status set" message over client connection,
        which sets the status of the specified account of the backend
        """

        msg = "account {} status set {}\r\n".format(account_id, status)
        logger.debug("sending status set message: %s", msg)
        self._send(msg)



"""
User Interface part of nuqql
"""

import curses
import curses.ascii
import datetime
import logging

from typing import TYPE_CHECKING, Any, Callable, Optional

import nuqql.config
import nuqql.conversation

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from nuqql.buddy import Buddy  # noqa

logger = logging.getLogger(__name__)


def handle_message(*args: Any) -> None:
    """
    Handle message from backend
    """

    # parse args
    backend, acc_id, chat, tstamp, sender, msg = args

    # convert timestamp
    tstamp = datetime.datetime.fromtimestamp(tstamp)

    # own message?
    own = False
    if sender == "<self>":
        own = True
        sender = "you"

    # look for an existing conversation and use it
    for conv in nuqql.conversation.CONVERSATIONS:
        if conv.backend is backend and \
           conv.account and conv.account.aid == acc_id and \
           conv.name == chat:
            # log message
            logger.debug("found conversation %s for message", conv.name)
            log_msg = conv.log(sender, msg, tstamp=tstamp, own=own)
            conv.history.log_to_file(log_msg)

            # if window is not already active notify user
            if not conv.is_input_win_active():
                conv.notify()
            return

    # nothing found, log to main window
    # or create temporary conversation
    account = backend.get_account(acc_id)
    if account is None:
        logger.debug("no conversation found for message")
        backend.conversation.log(sender, msg, tstamp=tstamp)
    else:
        # create buddy conversation
        conv = nuqql.conversation.BuddyConversation(backend, account, chat)
        conv.temporary = True
        conv.wins.list_win.add(conv)
        conv.wins.list_win.redraw_pad()
        logger.debug("created temporary conversation %s", conv.name)

        # log message
        log_msg = conv.log(sender, msg, tstamp=tstamp, own=own)
        conv.history.log_to_file(log_msg)

        # if window is not already active notify user
        if not conv.is_input_win_active():
            conv.notify()


def handle_chat_message(*args: Any) -> bool:
    """
    Handle chat message from backend
    """

    # parse args
    backend, acc_id, ctype, chat, nick, alias, status = args

    for conv in nuqql.conversation.CONVERSATIONS:
        if conv.backend is backend and \
           conv.account and conv.account.aid == acc_id and \
           conv.name == chat:
            # log chat message/event
            if alias == nick:
                log_msg = conv.log(chat, "{} {} [{}]".format(ctype, nick,
                                                             status))
            else:
                log_msg = conv.log(chat, "{} {} ({}) [{}]".format(
                    ctype, alias, nick, status))
            conv.history.log_to_file(log_msg)

            # if window is not already active notify user
            if not conv.is_input_win_active():
                conv.notify()
            logger.debug("handled chat message")
            return True

    # did not handle the message, return False
    logger.debug("chat message not handled")
    return False


def update_buddy(buddy: "Buddy") -> None:
    """
    Update buddy in UI
    """

    # look for existing buddy
    for conv in nuqql.conversation.CONVERSATIONS:
        if not isinstance(conv, nuqql.conversation.BuddyConversation):
            continue
        if conv.temporary:
            continue

        conv_buddy = conv.peers[0]
        if conv_buddy is buddy:
            conv.wins.list_win.redraw_pad()
            logger.debug("updated buddy %s in ui", buddy.name)
            return

    logger.debug("buddy %s not updated", buddy.name)


def add_buddy_to_temporary_conv(buddy: "Buddy") -> \
        Optional[nuqql.conversation.BuddyConversation]:
    """
    Try to find a temporary conversation for this buddy and add the buddy to it
    """

    for conv in nuqql.conversation.CONVERSATIONS:
        if not isinstance(conv, nuqql.conversation.BuddyConversation):
            continue
        if not conv.temporary:
            continue

        if conv.backend == buddy.backend and \
           conv.account == buddy.account and \
           conv.name == buddy.name:
            conv.peers.append(buddy)
            # conversation/buddy is now in backend, remove temporary flag
            conv.temporary = False
            logger.debug("added buddy %s to temporary conversation %s",
                         buddy.name, conv.name)
            return conv

    # nothing found, tell caller
    logger.debug("no temporary conversation found for buddy %s", buddy.name)
    return None


def add_buddy(buddy: "Buddy") -> None:
    """
    Add a new buddy to UI
    """

    # add new conversation for the buddy if necessary
    conv = add_buddy_to_temporary_conv(buddy)
    if not conv:
        # no temporary conversation found, add new one
        conv = nuqql.conversation.BuddyConversation(buddy.backend,
                                                    buddy.account,
                                                    buddy.name)
        conv.peers.append(buddy)
        conv.wins.list_win.add(conv)
        logger.debug("added conversation for buddy %s", buddy.name)

    # redraw list to show update
    conv.wins.list_win.redraw_pad()

    # check if there are unread messages for this new buddy in the history
    last_log_msg = conv.history.get_last_log_line()
    last_read_msg = conv.history.get_lastread()
    if last_log_msg:
        if not last_read_msg or not last_log_msg.is_equal(last_read_msg):
            # there are unread messages, notify user if
            # conversation is inactive
            if not conv.is_active():
                conv.notify()

    logger.debug("added buddy %s to ui", buddy.name)


def remove_buddy(buddy: "Buddy") -> None:
    """
    Remove a buddy from the UI
    """

    for index, conv in enumerate(nuqql.conversation.CONVERSATIONS):
        if not isinstance(conv, nuqql.conversation.BuddyConversation):
            continue
        if conv.temporary:
            # skip temporary conversations/buddies
            return

        conv_buddy = conv.peers[0]
        if conv_buddy is buddy:
            # if log window or input window are active, close them first
            if conv.is_log_win_active():
                # note: this activates the input window
                conv.wins.log_win.go_back()
            if conv.is_input_win_active():
                conv.wins.input_win.go_back()

            # remove conversation
            del nuqql.conversation.CONVERSATIONS[index]
            conv.wins.list_win.redraw()
            logger.debug("removed buddy %s from ui", buddy.name)
            return

    logger.debug("buddy %s not found/removed", buddy.name)


def read_input() -> str:
    """
    Read user input and return it to caller
    """

    # try to get input from user (timeout set in init())
    try:
        wch = nuqql.win.MAIN_WINS["screen"].get_wch()
    except curses.error:
        # no user input...
        wch = None

    return wch


def show_terminal_warning() -> None:
    """
    Show a warning that the terminal size is invalid, if it fits on screen
    """

    # clear terminal
    nuqql.win.MAIN_WINS["screen"].clear()

    # check if terminal is big enough for at least one character
    max_y, max_x = nuqql.win.MAIN_WINS["screen"].getmaxyx()
    if max_y < 1:
        return
    if max_x < 1:
        return

    # print as much of the error message as possible
    msg = "Invalid terminal size. Please resize."[:max_x - 1]
    nuqql.win.MAIN_WINS["screen"].addstr(0, 0, msg)
    logger.debug("showed terminal size invalid warning")


def is_input_valid(char: str) -> bool:
    """
    Helper that checks if input is valid
    """

    # is there a char at all?
    if char is None:
        return False

    # check for embedded 0 byte
    if char == "\0":
        return False

    return True


def handle_input() -> bool:
    """
    Read and handle user input
    """

    # list window is inactive -> user quit
    if not nuqql.win.MAIN_WINS["list"].state.active:
        return False

    # wait for user input and get timeout or character to process
    char = read_input()

    # handle user input
    if not is_input_valid(char):
        # No valid input, keep waiting for input
        return True

    # if terminal size is not valid, stop here
    if not nuqql.config.WinConfig.is_terminal_valid():
        show_terminal_warning()
        return True

    # if terminal resized, resize and redraw active windows
    if char == curses.KEY_RESIZE:
        nuqql.conversation.resize_main_window()
        return True

    # pass user input to active conversation
    for conv in nuqql.conversation.CONVERSATIONS:
        if conv.is_active():
            conv.process_input(char)
            return True

    # if no conversation is active pass input to active list window for
    # list window navigation
    nuqql.win.MAIN_WINS["list"].process_input(char)
    return True


def start(stdscr: Any, func: Callable) -> str:
    """
    Start UI and run provided function
    """

    logger.debug("starting ui")

    # save stdscr
    nuqql.win.MAIN_WINS["screen"] = stdscr

    # configuration
    stdscr.timeout(10)

    # clear everything
    stdscr.clear()
    stdscr.refresh()

    # disable cursor
    curses.curs_set(0)

    # make sure configs are loaded
    nuqql.config.init(stdscr)

    # create main windows, if terminal size is valid, otherwise just stop here
    if not nuqql.config.WinConfig.is_terminal_valid():
        logger.error("terminal size invalid")
        return "Terminal size invalid."
    nuqql.conversation.MainConversation(None, None, "").create_windows()
    nuqql.conversation.log_main_window(f"Welcome to nuqql v{nuqql.VERSION}!")
    nuqql.conversation.log_main_window("""Basic controls:
  * Navigate with arrow keys
  * Open conversation with <Enter>
  * Send message with <Ctrl-x>
  * Leave conversation with <Esc>
  * Exit nuqql with "q"
""")

    # run function provided by caller
    return func()


def init(func: Callable) -> None:
    """
    Initialize UI
    """

    logger.debug("initializing ui")
    retval = curses.wrapper(start, func)
    if retval and retval != "":
        print(retval)



"""
Nuqql conversation
"""

from .conversation import \
        Conversation, \
        CONVERSATIONS
from .helper import \
        log_main_window, \
        log_nuqql_conv, \
        remove_backend_conversations, \
        resize_main_window
from .buddyconversation import BuddyConversation
from .backendconversation import BackendConversation
from .nuqqlconversation import NuqqlConversation
from .mainconversation import MainConversation



"""
Backend conversation
"""

import logging

from typing import Tuple, TYPE_CHECKING

import nuqql.win

from .conversation import Conversation, CONVERSATIONS

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from nuqql.account import Account  # noqa
    from nuqql.backend import Backend  # noqa

logger = logging.getLogger(__name__)


class BackendConversation(Conversation):
    """
    Class for backend conversations
    """

    def create_windows(self) -> None:
        """
        Create windows for this conversation
        """

        logger.debug("creating windows for conversation %s", self.name)

        # create command windows for backends
        log_title = "Command log of {0}".format(self.name)
        input_title = "Command to {0}".format(self.name)
        self._create_windows_common(log_title, input_title)

        self.wins.list_win = nuqql.win.MAIN_WINS["list"]

    def get_name(self) -> str:
        """
        Get the name of the conversation, depending on type
        """

        notify = self._get_name_notification()
        return "{0}{{backend}} {1}".format(notify, self.name)

    def get_key(self) -> Tuple:
        """
        Get a key for sorting this conversation
        """

        # return tuple of sort keys:
        # notify, used, type, status, name
        return 0 - self.notification, 0, 1, 0, self.name

    @staticmethod
    def _check_chat_command(backend: "Backend", account: "Account", cmd: str,
                            name: str) -> None:
        """
        Check for chat commands
        """

        # check for existing conversation
        existing_conv = None
        existing_conv_index = -1
        for index, conv in enumerate(CONVERSATIONS):
            if not isinstance(conv, nuqql.conversation.BuddyConversation):
                continue
            if conv.backend == backend and \
               conv.account == account and \
               conv.name == name:
                existing_conv = conv
                existing_conv_index = index
                break

        # join: account <id> chat join <name>
        if cmd == "join":
            # make sure the conversation does not already exist
            if existing_conv:
                return

            # create temporary group conversation
            conv = nuqql.conversation.BuddyConversation(backend, account, name)
            conv.temporary = True
            conv.wins.list_win.add(conv)
            conv.wins.list_win.redraw()

        # part: account <id> chat part <name>
        if cmd == "part":
            # if it is (still) a temporary group conversation, remove it again
            if existing_conv and existing_conv.temporary:
                del CONVERSATIONS[existing_conv_index]
                existing_conv.wins.list_win.redraw()

    def _check_command(self, msg: str) -> None:
        """
        Check if msg is a special command we want to handle in nuqql before
        sending it to the backend
        """

        # TODO: test if we still need these functions

        parts = msg.split(" ")
        # check for account command
        # account <id> <command>
        if len(parts) < 3:
            return
        if parts[0] != "account":
            return

        # get account for this command
        assert self.backend
        account = self.backend.get_account(parts[1])
        if not account:
            return

        # check for delete command
        # account <id> delete
        if parts[2] == "delete":
            self.backend.delete_account(account.aid)
            return

        # check for chat commands
        # account <id> chat <join/part> <name>
        if len(parts) >= 5 and parts[2] == "chat":
            self._check_chat_command(self.backend, account, parts[3], parts[4])
            return

    def send_msg(self, msg: str) -> None:
        """
        Send message coming from the UI/input window
        """

        logger.debug("sending message %s in conversation %s", msg, self.name)

        self._send_msg_prepare(msg)

        # send command message to backend
        if self.backend and self.backend.client:
            # check for special commands to handle in nuqql first
            self._check_command(msg)

            self.backend.client.send_command(msg)

            if msg in ("bye", "quit"):
                # user told the backend to disconnect/quit, stop backend
                self.backend.stop()



"""
Nuqql UI Log Windows
"""

import logging

from types import SimpleNamespace
from typing import TYPE_CHECKING, Any, List

from .win import Win, MAIN_WINS

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from nuqql.config import WinConfig  # noqa
    from nuqql.conversation import Conversation  # noqa
    from nuqql.conversation.history import LogMessage  # noqa

logger = logging.getLogger(__name__)


class LogWin(Win):
    """
    Class for Log Windows
    """

    def __init__(self, config: "WinConfig", conversation: "Conversation",
                 title: str) -> None:
        Win.__init__(self, config, conversation, title)

        # list entries/message log
        self.list: List["LogMessage"] = []

        # string to search for
        self.search_input = ""
        self.search_text = ""
        self.search_keyfunc = {
            "DEL_CHAR":     self._process_search_input_del_char,
            "ENTER":        self._process_search_input_enter,
            "GO_BACK":      self._process_search_input_abort,
        }

        # user's view of the log
        self.view = SimpleNamespace(
            begin=-1,
            cur=-1
        )

    def add(self, entry: "LogMessage") -> None:
        """
        Add entry to internal list
        """

        # add entry to own list via base class function
        logger.debug("adding entry %s", entry)
        self.list_add(self.list, entry)

    def _get_log_view(self, props: SimpleNamespace) -> List["LogMessage"]:
        """
        Get a slice of the log in the current view of the pad
        """

        if self.view.begin == -1:
            start = len(self.list) - (props.win_size_y - props.pad_y_delta)
            start = max(start, 0)   # make sure it's >= 0
            end = len(self.list)
            self.view.cur = start
        else:
            start = self.view.begin
            end = props.win_size_y - props.pad_y_delta
            self.view.cur = start

        log_slice = []
        for index, msg in enumerate(self.list[start:]):
            if index > end:
                break
            log_slice.append(msg)

        return log_slice

    def _print_msg(self, msg: str, last: bool = False,
                   output: bool = True) -> int:
        """
        Print a single log message. Handle newlines in the log message and add
        additional line breaks for parts of the log message that are too long
        for the log window's pad width.
        """

        # get size of the pad
        max_y, max_x = self.pad.getmaxyx()

        # split the message at newlines and handle each line separately
        num_output = 0
        lines = msg.split("\n")
        for index, line in enumerate(lines):
            first = True
            while first or line:
                # always print the first part of a line, even if it's empty.
                first = False
                if output:
                    self.pad.insnstr(max_y - 1, 0, line, max_x)
                num_output += 1

                # drop all characters we printed from the current line
                line = line[max_x:]
                if not line and index == len(lines) - 1 and last:
                    # if we printed the last part of the last line of the last
                    # message, do not increase pad size and do not move cursor
                    break

                # there are more parts of the current line, more lines, or more
                # messages coming. Increase pad size and move cursor.
                if not output:
                    continue
                max_y += 1
                self.pad.resize(max_y, max_x)

        self.pad.move(max_y - 1, max_x - 1)
        return num_output

    def _print_log(self, props: SimpleNamespace) -> None:
        """
        dump log messages and resize pad according to new lines added
        """

        # get the current slice of the log and start with a fresh pad size
        log_slice = self._get_log_view(props)
        self.pad.resize(1, props.pad_size_x)

        for index, msg in enumerate(log_slice):
            # set colors and attributes for message:
            if not msg.own:
                # message from buddy
                if msg.is_read:
                    # old message
                    self.pad.attrset(self.config.attr["log_win_text_peer_old"])
                else:
                    # new message
                    self.pad.attrset(self.config.attr["log_win_text_peer_new"])
            else:
                # message from you
                if msg.is_read:
                    # old message
                    self.pad.attrset(self.config.attr["log_win_text_self_old"])
                else:
                    # new message
                    self.pad.attrset(self.config.attr["log_win_text_self_new"])

            # output message
            if index < len(log_slice) - 1:
                self._print_msg(msg.read())
            else:
                self._print_msg(msg.read(), last=True)

    def _get_properties(self) -> SimpleNamespace:
        """
        Get window/pad properties, depending on max size and zoom
        """

        props = SimpleNamespace()
        props.max_y, props.max_x = MAIN_WINS["screen"].getmaxyx()
        if self.state.zoomed:
            # window is currently zoomed
            props.pos_y, props.pos_x = 0, 0
            props.pos_y_off, props.pos_x_off = 1, 0
            props.win_size_y, props.win_size_x = props.max_y, props.max_x
            props.pad_size_y, props.pad_size_x = (props.max_y - 2, props.max_x)
            props.pad_y_delta, props.pad_x_delta = 2, 0
        else:
            # window is not zoomed
            props.pos_y, props.pos_x = self.config.get_pos()
            props.pos_y_off, props.pos_x_off = 1, 1
            props.win_size_y, props.win_size_x = self.config.get_size()
            # use actual pad size, it will be resized later if necessary
            props.pad_size_y, props.pad_size_x = self.pad.getmaxyx()
            props.pad_y_delta, props.pad_x_delta = 2, 2

        return props

    def _pad_refresh(self, props: SimpleNamespace) -> None:
        """
        Helper for running move_pad(), check_borders(), and pad.refresh()
        """
        self._move_pad()
        self._check_borders()
        self.pad.refresh(self.state.pad_y, self.state.pad_x,
                         props.pos_y + props.pos_y_off,
                         props.pos_x + props.pos_x_off,
                         props.pos_y + props.win_size_y - props.pad_y_delta,
                         props.pos_x + props.win_size_x - props.pad_x_delta)

    def redraw_pad(self) -> None:
        # if terminal size is invalid, stop here
        if not self.config.is_terminal_valid():
            return

        # screen/pad properties
        props = self._get_properties()

        # if window was resized, resize pad size according to new window size
        if props.pad_size_x != props.win_size_x - props.pad_x_delta:
            props.pad_size_x = props.win_size_x - props.pad_x_delta
            self.pad.resize(props.pad_size_y, props.pad_size_x)
        if props.pad_size_y != props.win_size_y - props.pad_y_delta:
            props.pad_size_y = props.win_size_y - props.pad_y_delta
            self.pad.resize(props.pad_size_y, props.pad_size_x)
            self.state.pad_y = 0  # reset pad position

        # print log
        self.pad.erase()
        self._print_log(props)

        # check if visible part of pad needs to be moved and display it
        self.state.cur_y, self.state.cur_x = self.pad.getyx()
        self._pad_refresh(props)

    def _cursor_top(self, *args: Any) -> None:
        # jump to first line in log
        logger.debug("move cursor to top of log")
        if self.view.cur > 0:
            # view is not at the top yet, so move it there
            self.view.begin = 0
            self.redraw_pad()

        # move cursor to top
        self.state.cur_y, self.state.cur_x = 0, 0
        self.pad.move(self.state.cur_y, self.state.cur_x)
        props = self._get_properties()
        self._pad_refresh(props)

    def _cursor_bottom(self, *args: Any) -> None:
        # jump to last line in log
        logger.debug("move cursor to bottom of log")
        props = self._get_properties()
        view_size = props.win_size_y - props.pad_y_delta

        if self.view.cur < len(self.list) - view_size:
            # view is not at the bottom yet, so move it there
            self.view.begin = len(self.list) - view_size
            self.redraw_pad()

        # move cursor to bottom
        self.state.cur_y, self.state.cur_x = self.pad.getmaxyx()[0] - 1, 0
        self.pad.move(self.state.cur_y, self.state.cur_x)
        props = self._get_properties()
        self._pad_refresh(props)

    def _cursor_page_up(self, *args: Any) -> None:
        # move cursor up one page until first entry in log
        logger.debug("move cursor up one page")
        props = self._get_properties()
        view_size = props.win_size_y - props.pad_y_delta

        cur_y, _cur_x = self.pad.getyx()
        lines_left = view_size - cur_y
        if lines_left <= 0:
            # we can stay in the current view
            self.state.cur_y = 0 - lines_left
            self.state.cur_x = 0
        else:
            # we need to get more messages
            while self.view.cur > 0:
                self.view.begin = self.view.cur - 1

                # get next message and see if we have enough lines
                log_slice = self._get_log_view(props)
                num_lines = self._print_msg(log_slice[0].read(), output=False)
                lines_left -= num_lines
                if lines_left <= 0:
                    # we got all missing lines
                    break

            # show updated log_view in pad and set cursor position
            self.redraw_pad()
            self.state.cur_y = max(0, 0 - lines_left)
            self.state.cur_x = 0

        # move cursor up to previously determined position
        self.pad.move(self.state.cur_y, self.state.cur_x)
        props = self._get_properties()
        self._pad_refresh(props)

    def _cursor_page_down(self, *args: Any) -> None:
        # move cursor down one page until last entry in log
        logger.debug("move cursor down one page")
        props = self._get_properties()
        view_size = props.win_size_y - props.pad_y_delta

        cur_y, _cur_x = self.pad.getyx()
        max_y, _max_x = self.pad.getmaxyx()
        lines_left = view_size - (max_y - 1 - cur_y)
        if lines_left <= 0:
            # we can stay in the current view
            self.state.cur_y = max_y - 1 + lines_left
            self.state.cur_x = 0
        else:
            # we need to get more messages
            while self.view.cur < len(self.list) - view_size:
                self.view.begin = self.view.cur + 1

                # get next message and see if we have enough lines
                log_slice = self._get_log_view(props)
                num_lines = self._print_msg(log_slice[-1].read(), output=False)
                lines_left -= num_lines
                if lines_left <= 0:
                    # we got all missing lines
                    break

            # show updated log_view in pad and set cursor position
            self.redraw_pad()
            self.state.pad_y = 0    # make sure we only show up to first line
            max_y, _max_x = self.pad.getmaxyx()
            self.state.cur_y = min(max_y - 1, max_y - 1 + lines_left)
            self.state.cur_x = 0

        # move cursor down to previously determined position
        self.pad.move(self.state.cur_y, self.state.cur_x)
        props = self._get_properties()
        self._pad_refresh(props)

    def _cursor_up(self, *args: Any) -> None:
        # move cursor up until first entry in list
        logger.debug("move cursor up")
        if self.state.cur_y > 0:
            # inside current view, simply move cursor up
            self.state.cur_y, self.state.cur_x = self.state.cur_y - 1, 0

        elif self.view.cur > 0:
            # at top of current view, move view up
            self.view.begin = self.view.cur - 1

            # if the previous message is multi line, only go to last line
            props = self._get_properties()
            log_slice = self._get_log_view(props)
            num_lines = self._print_msg(log_slice[0].read(), output=False)
            self.redraw_pad()
            self.state.cur_y, self.state.cur_x = num_lines - 1, 0

        # move cursor up
        self.pad.move(self.state.cur_y, self.state.cur_x)
        props = self._get_properties()
        self._pad_refresh(props)

    def _cursor_down(self, *args: Any) -> None:
        # move cursor down until end of list
        logger.debug("move cursor down")
        props = self._get_properties()
        lines = self.pad.getmaxyx()[0] - 1
        view_size = props.win_size_y - props.pad_y_delta

        if self.state.cur_y < lines:
            # inside current view, simply move cursor down
            self.state.cur_y, self.state.cur_x = self.state.cur_y + 1, 0

        elif self.view.cur < len(self.list) - view_size:
            # at bottom of current view, move view down
            self.view.begin = self.view.cur + 1

            # if the next message is multi line, only go to first line
            props = self._get_properties()
            log_slice = self._get_log_view(props)
            num_lines = self._print_msg(log_slice[-1].read(), output=False)
            self.redraw_pad()
            self.state.pad_y = 0    # make sure we only show up to first line
            self.state.cur_y, self.state.cur_x = \
                self.pad.getmaxyx()[0] - num_lines, 0

        # move cursor down
        self.pad.move(self.state.cur_y, self.state.cur_x)
        props = self._get_properties()
        self._pad_refresh(props)

    def zoom_win(self) -> None:
        """
        Zoom in and out of log window
        """

        # get positions and sizes for zoomed and normal mode
        if self.state.zoomed:
            logger.debug("zooming out")
            self.state.zoomed = False
        else:
            logger.debug("zooming in")
            self.state.zoomed = True
        props = self._get_properties()

        # resize window and pad
        self.resize_win(props.win_size_y, props.win_size_x)
        self.move_win(props.pos_y, props.pos_x)
        self.pad.resize(props.win_size_y - props.pad_y_delta,
                        props.win_size_x - props.pad_x_delta)
        self._move_pad()
        self._check_borders()

        # redraw everything
        if self.state.zoomed:
            # make list win and input win invisible
            self.conversation.wins.list_win.state.visible = False
            self.conversation.wins.input_win.state.visible = False

            # redraw window
            self.redraw()
        else:
            # make list win and input win visible
            self.conversation.wins.list_win.state.visible = True
            self.conversation.wins.input_win.state.visible = True

            # redraw everything
            MAIN_WINS["screen"].clear()
            MAIN_WINS["screen"].refresh()
            self.conversation.wins.list_win.redraw()
            self.conversation.wins.log_win.redraw()
            self.conversation.wins.input_win.redraw()

    def go_back(self, *_args: Any) -> None:
        """
        Go back to input win
        """

        logger.debug("leaving window")

        # if window was zoomed, switch back to normal view
        if self.state.zoomed:
            self.zoom_win()

        # reactivate input window
        self.state.active = True
        self.conversation.wins.input_win.state.active = True

        # show last messages
        self.view.begin = -1

        # redraw pad to display messages received in the meantime...
        self.redraw_pad()

        # ...and clear notifications for these messages
        self.conversation.clear_notifications()

    def _search(self, *args: Any) -> None:
        """
        Search: Start a new search dialog
        """

        logger.debug("starting search input mode")

        # enter search input mode
        self.search_input = "/"
        self._process_search_input_show()

    def _process_search_input_show(self) -> None:
        """
        Show current search string
        """

        if not self.search_input:
            # show window border again
            self.redraw()
            return

        # TODO: use config get_size or something?
        max_y, max_x = self.win.getmaxyx()
        show = self.search_input.ljust(max_x - 4, " ")
        self.win.addnstr(max_y - 1, 2, show, max_x - 4)
        self.win.refresh()

    def _process_search_input_del_char(self) -> None:
        """
        Delete char from search input field
        """

        if not self.search_input:
            return

        logger.debug("deleting char from search input")
        self.search_input = self.search_input[:-1]
        self._process_search_input_show()

    def _process_search_input_enter(self) -> None:
        """
        Search for string in search input field
        """

        if not self.search_input:
            return

        self.search_text = self.search_input[1:]
        logger.debug("starting search for %s", self.search_text)
        self.search_input = ""
        self._process_search_input_show()
        self._search_next()

    def _process_search_input_abort(self) -> None:
        """
        Abort search
        """

        logger.debug("leaving search input mode")
        self.search_input = ""
        self._process_search_input_show()

    def _search_next(self, *args: Any) -> None:
        """
        Search for next match
        """

        # skip this if we are not in search mode
        if self.search_text == "":
            return

        # init
        props = self._get_properties()

        # search views for text until first view
        while self.view.cur >= 0:
            # search current view for text until first line
            while self.state.cur_y >= 0:
                _cur_text = self.pad.instr(self.state.cur_y, 0)
                cur_text = _cur_text.decode()   # type: ignore
                cur_text = cur_text[:self.state.cur_x]
                index = cur_text.rfind(self.search_text)
                if index != -1:
                    # found it, stop here
                    self.pad.move(self.state.cur_y, index)
                    self.state.cur_y, self.state.cur_x = self.pad.getyx()

                    self._pad_refresh(props)
                    return

                # keep searching in next line
                self.state.cur_y -= 1
                self.state.cur_x = props.pad_size_x

            # reached end of view, move view further up
            if self.view.cur == 0:
                # reached top already
                break
            self._cursor_up()
            self.state.cur_x = self.pad.getmaxyx()[1] - 1

        # make sure current cursor position is visible even if nothing found
        self._pad_refresh(props)

    def search_next(self) -> None:
        """
        Helper for calling search from other windows
        """

        # make sure we find something on the current line
        logger.debug("searching for next match")
        self.state.cur_x = self.pad.getmaxyx()[1]
        self._search_next()

    def _search_prev(self, *args: Any) -> None:
        """
        Search for previous match
        """

        # skip this if we are not in search mode
        if self.search_text == "":
            return

        logger.debug("searching for previous match")

        # init
        props = self._get_properties()
        view_size = props.win_size_y - props.pad_y_delta

        # if we are already on a match, skip it
        _cur_text = self.pad.instr(self.state.cur_y, 0)
        cur_text = _cur_text.decode()   # type: ignore
        cur_text = cur_text[self.state.cur_x:]
        cur_text = cur_text[:len(self.search_text)]
        if cur_text == self.search_text:
            self.state.cur_x += len(self.search_text)

        # search views for text until last view
        while self.view.cur <= max(0, len(self.list) - view_size):
            # search current view for text until end of view
            while self.state.cur_y <= self.pad.getmaxyx()[0]:
                _cur_text = self.pad.instr(self.state.cur_y, 0)
                cur_text = _cur_text.decode()   # type: ignore
                cur_text = cur_text[self.state.cur_x:]
                index = cur_text.find(self.search_text)

                # found it, stop here
                if index != -1:
                    self.pad.move(self.state.cur_y, self.state.cur_x + index)
                    self.state.cur_y, self.state.cur_x = self.pad.getyx()

                    self._pad_refresh(props)
                    return

                # keep searching
                self.state.cur_y += 1
                self.state.cur_x = 0    # set it to first position in line

            # reached end of view, move view further down
            if self.view.cur == max(0, len(self.list) - view_size):
                # reached bottom already
                pad_size_y, pad_size_x = self.pad.getmaxyx()
                self.pad.move(pad_size_y - 1, pad_size_x - 1)
                self.state.cur_y, self.state.cur_x = self.pad.getyx()
                break
            self._cursor_down()

        # make sure current cursor position is visible even if nothing found
        self._pad_refresh(props)

    def process_input(self, char: str) -> None:
        """
        Process user input
        """

        self.state.cur_y, self.state.cur_x = self.pad.getyx()

        # check if we are in search input mode
        if self.search_input:
            # search input: look for special key mapping or process as text
            if not self.handle_keybinds(
                    char, keybinds=self.config.keybinds["__search__"],
                    keyfunc=self.search_keyfunc):
                # no special key, add character to search input field
                try:
                    self.search_input += char
                    self._process_search_input_show()
                except (ValueError, TypeError):
                    pass
        else:
            # look for special key mappings in keymap
            self.handle_keybinds(char)

        # display changes in the pad
        # TODO: switch this back on and remove redraw code from other methods?
        # self.redraw_pad()



"""
Nuqql account
"""

import logging
import time

from typing import List, TYPE_CHECKING

import nuqql.ui

from nuqql.buddy import Buddy

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from nuqql.backend import Backend

logger = logging.getLogger(__name__)

# update buddies only every BUDDY_UPDATE_TIMER seconds
BUDDY_UPDATE_TIMER = 5


class Account:
    """
    Class for Accounts
    """

    def __init__(self, aid: str, prot: str, user: str) -> None:
        self.aid = aid
        self.name = user
        self.type = prot
        self.buddies: List[Buddy] = []
        self.buddies_update = 0
        logger.debug("created new account: aid %s, name %s, type %s",
                     self.aid, self.name, self.type)

    def update_buddies(self) -> bool:
        """
        Update the buddy list of this account.
        Return True if an update is pending, False otherwise.
        """

        # update only once every BUDDY_UPDATE_TIMER seconds
        if time.time() - self.buddies_update <= BUDDY_UPDATE_TIMER:
            return False
        self.buddies_update = int(time.time())

        # remove buddies, that have not been updated for a while
        for rem in [buddy for buddy in self.buddies if not buddy.updated]:
            nuqql.ui.remove_buddy(rem)
            logger.debug("removed buddy %s from account %s on backend %s",
                         rem.name, self.aid, rem.backend.name)
        self.buddies = [buddy for buddy in self.buddies if buddy.updated]

        # set update pending in buddy
        for buddy in self.buddies:
            buddy.updated = False

        return True

    def update_buddy(self, backend: "Backend", name: str, alias: str,
                     status: str) -> None:
        """
        Update a single buddy of this account. Could be a new buddy.
        """

        # look for existing buddy
        for buddy in self.buddies:
            if buddy.name == name:
                if buddy.update(status, alias):
                    # tell ui about the update
                    nuqql.ui.update_buddy(buddy)

                # found existing buddy; stop here
                return

        # new buddy
        new_buddy = Buddy(backend, self, name)
        new_buddy.update(status, alias)
        self.buddies.append(new_buddy)
        logger.debug("added new buddy %s to account %s on backend %s",
                     name, self.aid, backend.name)

        # tell ui there is a new buddy
        nuqql.ui.add_buddy(new_buddy)

    def flush_buddies(self) -> None:
        """
        Flush all buddies of this account
        """

        for buddy in self.buddies:
            nuqql.ui.remove_buddy(buddy)
        self.buddies = []



"""
Backend class
"""

import logging
import socket
import time

from pathlib import Path
from typing import Dict, Optional, Tuple

import nuqql.conversation
import nuqql.ui

from nuqql.account import Account
from .server import BackendServer
from .client import BackendClient
from .parse import parse_msg

logger = logging.getLogger(__name__)


class Backend:
    """
    Class for backends. Allows starting server processes and connecting to
    (self-started or externally started) servers
    """

    def __init__(self, name: str) -> None:
        # backend
        self.name = name
        self.accounts: Dict[str, Account] = {}
        # conversation for communication with the backend.
        self.conversation: Optional[nuqql.conversation.Conversation] = None

        # all backends
        self.backends: Dict[str, Backend] = {}

        # server
        self.server: Optional[BackendServer] = None

        # client
        self.client: Optional[BackendClient] = None

        # self.collect_acc = -1

    def start_server(self, cmd: str, path: str) -> None:
        """
        Add a server to this backend and start it
        """

        logger.debug("starting server of backend %s: cmd: %s, path: %s",
                     self.name, cmd, path)
        self.server = BackendServer(cmd, path)
        self.server.start()

    def stop_server(self) -> None:
        """
        Stop the server of this backend
        """

        logger.debug("stopping server of backend %s", self.name)
        if self.server:
            self.server.stop()

    def start_client(self) -> None:
        """
        Start the backend's client
        """

        logger.debug("starting client of backend %s", self.name)
        if self.client:
            self.client.start()

    def init_client(self, sock_af: "socket.AddressFamily" = socket.AF_UNIX,
                    ip_addr: str = "127.0.0.1", port: int = 32000,
                    sock_file: str = "") -> None:
        """
        Add a client to this backend
        """

        logger.debug("initializing client of backend %s: "
                     "sock_af: %s, ip_addr: %s, port: %s, sock_file: %s",
                     self.name, sock_af, ip_addr, port, sock_file)
        self.client = BackendClient(sock_af, ip_addr, port, sock_file)
        self.client.backend = self

    def stop_client(self) -> None:
        """
        Stop the client of this backend
        """

        logger.debug("stopping client of backend %s", self.name)
        if self.client:
            self.client.stop()

    def _handle_network(self, msg) -> None:
        # parse it
        parsed_msg = parse_msg(msg)
        msg_type = parsed_msg[0]

        # handle info message or error message
        if msg_type in ("info", "error"):
            text = msg_type + ": " + parsed_msg[1]
            if self.conversation:
                self.conversation.log("nuqql", text)
            return

        # handle account message
        if msg_type == "account":
            self.handle_account_msg(parsed_msg)
            return

        # handle status message
        if msg_type == "status":
            text = "account {} status: {}".format(parsed_msg[1], parsed_msg[2])
            if self.conversation:
                self.conversation.log("nuqql", text)
            return

        # handle chat message
        if msg_type == "chat":
            self.handle_chat_msg(parsed_msg)
            return

        # handle buddy messages
        if msg_type == "buddy":
            self.handle_buddy_msg(parsed_msg)
            return

        # handle normal messages and error messages
        # TODO: handle error messages somewhere else?
        if msg_type in ("message", "parsing error"):
            self.handle_message_msg(parsed_msg)

    def handle_network(self) -> None:
        """
        Try to read from the client connection and handle messages.
        """

        # try to read message
        if not self.client:
            return
        msg = self.client.read()
        if msg is None:
            return

        logger.debug("handling message from network in backend %s: %s",
                     self.name, msg)
        self._handle_network(msg)

    def _parse_message_account_specific(self, acc_id: str, sender: str,
                                        parsed_msg: Tuple[str, ...]
                                        ) -> Tuple[str, str]:
        msg = parsed_msg[5]
        resource = sender
        for tmp_acc in self.accounts.values():
            if tmp_acc.aid == acc_id:
                if tmp_acc.type == "icq":
                    if sender[-1] == ":":
                        sender = sender[:-1]
                    if msg[:6] == "<BODY>":
                        msg = msg[6:]
                    if msg[-7:] == "</BODY>":
                        msg = msg[:-7]
                    break
                if tmp_acc.type == "xmpp":
                    sender_parts = sender.split("/")
                    sender = sender_parts[0]
                    if len(sender_parts) > 1:
                        resource = sender_parts[1]
                    break
                if tmp_acc.type == "matrix":
                    if sender != "<self>":
                        resource = sender[1:].split(":")[0]
                    sender = parsed_msg[2]
                    break
        return sender, resource

    def handle_message_msg(self, parsed_msg: Tuple[str, ...]) -> None:
        """
        Handle "message" message
        """

        logger.debug("handling message in backend %s", self.name)

        # msg_type = parsed_msg[0]
        acc_id = parsed_msg[1]
        # destination = parsed_msg[2]
        tstamp = parsed_msg[3]
        sender = parsed_msg[4]
        msg = parsed_msg[5]

        # account specific message parsing
        sender, _resource = self._parse_message_account_specific(
            acc_id, sender, parsed_msg)

        # let ui handle the message
        nuqql.ui.handle_message(self, acc_id, sender, tstamp, sender, msg)

    def handle_chat_msg(self, parsed_msg: Tuple[str, ...]) -> None:
        """
        Handle Chat message
        """

        logger.debug("handling chat message in backend %s", self.name)

        # "chat", ctype, acc, chat, nick
        ctype = parsed_msg[1]
        acc_id = parsed_msg[2]
        chat = parsed_msg[3]

        # msg message
        if ctype == "msg:":
            timestamp = parsed_msg[4]
            sender = parsed_msg[5]
            msg = parsed_msg[6]

            # account specific message parsing, use resource as sender
            _, sender = self._parse_message_account_specific(
                acc_id, sender, parsed_msg)

            # handle message in ui
            nuqql.ui.handle_message(self, acc_id, chat, timestamp, sender, msg)
            return

        # user message
        if ctype == "user:":
            nick = parsed_msg[4]
            alias = parsed_msg[5]
            status = parsed_msg[6]
            # if there is a conversation for this type and group chat, log to
            # it. Otherwise, just log to backend conversation later
            if nuqql.ui.handle_chat_message(self, acc_id, ctype, chat, nick,
                                            alias, status):
                return

        # list message
        if ctype == "list:":
            chat_alias = parsed_msg[4]
            nick = parsed_msg[5]
            if chat != chat_alias:
                chat = "{} ({})".format(chat_alias, chat)

        # log to backend conversation
        text = "account {} chat: {} {} {}".format(acc_id, ctype, chat, nick)
        if self.conversation:
            self.conversation.log("nuqql", text)

    def handle_account_msg(self, parsed_msg: Tuple[str, ...]) -> None:
        """
        Handle Account message
        """

        logger.debug("handling account message in backend %s", self.name)

        # "account", acc_id, acc_alias, acc_prot, acc_user, acc_status
        # msg_type = parsed_msg[0]
        acc_id = parsed_msg[1]
        acc_alias = parsed_msg[2]
        acc_prot = parsed_msg[3]
        acc_user = parsed_msg[4]
        acc_status = parsed_msg[5]

        # output account
        text = "account {0} ({1}) {2} {3} {4}.".format(acc_id, acc_alias,
                                                       acc_prot, acc_user,
                                                       acc_status)
        if self.conversation:
            self.conversation.log("nuqql", text)

        if not self.client:
            return

        # do not add account if it already exists
        if acc_user in self.accounts:
            return

        # new account, add it
        acc = Account(acc_id, acc_prot, acc_user)
        self.accounts[acc.name] = acc

        # collect buddies from backend
        text = "Collecting buddies for {0} account {1}: {2}.".format(
            acc.type, acc.aid, acc.name)
        if self.conversation:
            self.conversation.log("nuqql", text)
        acc.buddies_update = int(time.time())
        self.client.send_buddies(acc.aid)

        # collect messages from backend
        text = "Collecting messages for {0} account {1}: {2}.".format(
            acc.type, acc.aid, acc.name)
        if self.conversation:
            self.conversation.log("nuqql", text)
        self.client.send_collect(acc.aid)

        # if there is a global_status, set account status to it
        status = self.read_global_status()
        if status != "":
            self.client.send_status_set(acc_id, status)

    def handle_buddy_msg(self, parsed_msg: Tuple[str, ...]) -> None:
        """
        Handle Buddy message
        """

        logger.debug("handling buddy message in backend %s", self.name)

        # get message parts
        # msg_type = parsed_msg[0]
        acc_id = parsed_msg[1]
        status = parsed_msg[2]
        name = parsed_msg[3]
        alias = parsed_msg[4]

        # if there is no alias, just use name
        if alias == "":
            alias = name

        # handle buddy update
        for account in self.accounts.values():
            if account.aid == acc_id:
                account.update_buddy(self, name, alias, status)
                return

    def update_buddies(self) -> None:
        """
        Update buddies of this account
        """

        if not self.client:
            return

        # update buddies
        for acc in self.accounts.values():
            if acc.update_buddies():
                logger.debug("updating buddies in backend %s", self.name)
                self.client.send_buddies(acc.aid)

    def get_account(self, account_id: str) -> Optional["Account"]:
        """
        Get account with specified account id
        """

        logger.debug("getting account in backend %s", self.name)
        for acc in self.accounts.values():
            if acc.aid == account_id:
                return acc

        return None

    def delete_account(self, account_id: str) -> None:
        """
        Delete account identified by the account id
        """

        account = self.get_account(account_id)
        if account:
            logger.debug("removing account %s in backend %s", account_id,
                         self.name)
            account.flush_buddies()
            del self.accounts[account.name]

    def read_global_status(self) -> str:
        """
        Read global status from global_status file
        """

        logger.debug("reading global status in backend %s", self.name)

        # if there is a global_status file, read it
        global_status_dir = str(Path.home()) + "/.config/nuqql"
        Path(global_status_dir).mkdir(parents=True, exist_ok=True)
        global_status_file = global_status_dir + "/global_status"
        try:
            with open(global_status_file) as status_file:
                line = status_file.readline()
                status = line.split()
                if not status:
                    return ""
                return status[0]
        except FileNotFoundError:
            return ""

    def stop(self) -> None:
        """
        Stop the backend, Note: changes BACKENDS
        """

        logger.debug("stopping backend %s", self.name)

        # print to main window
        log_msg = "Stopping client and server for backend \"{0}\".".format(
            self.name)
        nuqql.conversation.log_nuqql_conv(log_msg)

        # stop client and server
        self.stop_client()
        self.stop_server()

        # remove backend from backends dict
        del self.backends[self.name]  # changes BACKENDS, be carefull

        # remove conversation and update list window
        nuqql.conversation.remove_backend_conversations(self)



"""
Nuqql backend code
"""

from .backend import Backend
from .nuqqlbackend import NuqqlBackend
from .helpers import start_backends, update_buddies, handle_network, \
    stop_backends
from .server import BackendServer
from .client import BackendClient



"""
nuqql conversation helpers
"""

import datetime
import logging

from typing import TYPE_CHECKING

import nuqql.win

from .conversation import CONVERSATIONS
from .logmessage import LogMessage

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from nuqql.backend import Backend  # noqa

logger = logging.getLogger(__name__)


def remove_backend_conversations(backend: "Backend") -> None:
    """
    Remove all conversations beloning to the backend
    """

    logger.debug("removing all conversations of backend %s", backend.name)
    for conv in CONVERSATIONS[:]:
        if conv.backend == backend:
            CONVERSATIONS.remove(conv)
            conv.wins.list_win.redraw()
            logger.debug("removed conversation %s of backend %s",
                         conv.name, backend.name)


def log_main_window(msg: str) -> None:
    """
    Log message to main windows
    """

    logger.debug("logging message to main window: %s", msg)
    now = datetime.datetime.now()
    log_msg = LogMessage(now, "nuqql", msg)
    nuqql.win.MAIN_WINS["log"].add(log_msg)


def log_nuqql_conv(msg: str) -> None:
    """
    Log message to the nuqql conversation
    """

    logger.debug("logging message to nuqql conversation: %s", msg)
    for conv in CONVERSATIONS:
        if conv.name == "nuqql":
            conv.log("nuqql", msg)
            return


def resize_main_window() -> None:
    """
    Resize main window
    """

    logger.debug("resizing main window")

    # get main win
    screen = nuqql.win.MAIN_WINS["screen"]

    # get new maxima
    max_y, max_x = screen.getmaxyx()

    # redraw main windows
    screen.clear()
    screen.refresh()

    # redraw conversation windows
    found_active = False
    for conv in CONVERSATIONS:
        # resize and move conversation windows
        if conv.wins.list_win:
            size_y, size_x = conv.wins.list_win.config.get_size()
            conv.wins.list_win.resize_win(size_y, size_x)
        if conv.wins.log_win:
            # TODO: move zoom/resizing to win.py?
            if conv.wins.log_win.state.zoomed:
                size_y, size_x = max_y, max_x
                pos_y, pos_x = 0, 0
                conv.wins.log_win.state.pad_y = 0  # reset pad position
            else:
                size_y, size_x = conv.wins.log_win.config.get_size()
                pos_y, pos_x = conv.wins.log_win.config.get_pos()
            conv.wins.log_win.resize_win(size_y, size_x)
            conv.wins.log_win.move_win(pos_y, pos_x)
        if conv.wins.input_win:
            size_y, size_x = conv.wins.input_win.config.get_size()
            conv.wins.input_win.resize_win(size_y, size_x)
            pos_y, pos_x = conv.wins.input_win.config.get_pos()
            conv.wins.input_win.move_win(pos_y, pos_x)
        # redraw active conversation windows
        if conv.is_active():
            found_active = True
            conv.wins.list_win.redraw()
            conv.wins.input_win.redraw()
            conv.wins.log_win.redraw()

    # if there are no active conversations, redraw nuqql main windows
    if not found_active:
        # list win
        list_win = nuqql.win.MAIN_WINS["list"]
        size_y, size_x = list_win.config.get_size()

        list_win.resize_win(size_y, size_x)
        list_win.redraw()

        # log main win
        log_win = nuqql.win.MAIN_WINS["log"]
        size_y, size_x = log_win.config.get_size()
        pos_y, pos_x = log_win.config.get_pos()

        log_win.resize_win(size_y, size_x)
        log_win.move_win(pos_y, pos_x)
        log_win.redraw()



"""
Backend helpers
"""

import logging
import shutil
import os
import time

from pathlib import Path
from typing import Dict, List, Optional

import nuqql.config
import nuqql.conversation
from nuqql import VERSION

from .backend import Backend
from .nuqqlbackend import NuqqlBackend

logger = logging.getLogger(__name__)

# how long should we wait for backends (in seconds) before starting clients
BACKENDS_WAIT_TIME = 1

# disable the python backends' own history?
BACKEND_DISABLE_HISTORY = True

# let backend push accounts to us?
BACKEND_PUSH_ACCOUNTS = True

# filenames that should not get started as backends
BACKEND_BLACKLIST = ["nuqql-keys", "nuqql-based"]

# dictionary for all active backends
BACKENDS: Dict[str, "Backend"] = {}


def update_buddies() -> None:
    """
    Helper for updating buddies on all backends
    """

    for backend in dict(BACKENDS).values():
        backend.update_buddies()


def handle_network() -> None:
    """
    Helper for handling network events on all backends
    """

    for backend in dict(BACKENDS).values():
        backend.handle_network()


def start_backend(backend_name: str, backend_exe: str, backend_path: str,
                  backend_cmd_fmt: str,
                  backend_sockfile: str) -> Optional[Backend]:
    """
    Helper for starting a backend
    """

    logger.debug("starting backend: name %s, exe %s, path %s, cmd_fmt %s, "
                 "sockfile %s", backend_name, backend_exe, backend_path,
                 backend_cmd_fmt, backend_sockfile)

    # check if backend exists in path
    exe = shutil.which(backend_exe, path=os.getcwd())
    if exe is None:
        exe = shutil.which(backend_exe)
    if exe is None:
        # does not exist, stop here
        return None

    backend_cmd = backend_cmd_fmt.format(exe, backend_path)

    backend = Backend(backend_name)
    backend.backends = BACKENDS
    backend.start_server(cmd=backend_cmd, path=backend_path)
    backend.init_client(sock_file=backend_sockfile)

    BACKENDS[backend_name] = backend

    # add conversation and show it in list window
    conv = nuqql.conversation.BackendConversation(backend, None, backend.name)
    conv.create_windows()
    nuqql.conversation.CONVERSATIONS.append(conv)
    backend.conversation = conv
    conv.wins.list_win.redraw()

    # log backend start
    log_msg = "Starting backend \"{0}\".".format(backend.name)
    nuqql.conversation.log_nuqql_conv(log_msg)

    # return the backend
    return backend


def start_backend_from_path(filename) -> Optional[Backend]:
    """
    Helper for starting a single backend found in PATH.
    """

    logger.debug("starting backend found in PATH")

    backend_name = _get_backend_name(filename)
    backend_exe = filename
    backend_path = str(Path.home()) + f"/.config/nuqql/backend/{backend_name}"
    backend_cmd_fmt = "{0} --af unix --dir {1} --sockfile " + \
        f"{backend_name}.sock"
    if BACKEND_DISABLE_HISTORY:
        backend_cmd_fmt += " --disable-history"
    if BACKEND_PUSH_ACCOUNTS:
        backend_cmd_fmt += " --push-accounts"
    loglevel = nuqql.config.get("loglevel")
    if loglevel:
        backend_cmd_fmt += f" --loglevel {loglevel}"
    backend_sockfile = backend_path + f"/{backend_name}.sock"

    return start_backend(backend_name, backend_exe, backend_path,
                         backend_cmd_fmt, backend_sockfile)


def _is_backend_filename(filename: str) -> bool:
    """
    Check if given filename matches a backend name
    """

    if filename.startswith("nuqql-"):
        return True
    if filename == "purpled":
        return True
    return False


def _get_backend_name(filename: str) -> str:
    """
    Return the backend name for given backend filename
    """

    if filename == "purpled":
        return filename
    return filename[6:]


def get_backends_from_path() -> List[str]:
    """
    Get a list of backends found in PATH.
    """

    backends: List[str] = []
    for path_dir in os.get_exec_path():
        if not os.path.isdir(path_dir):
            # PATH may contain directories that do not exist, skip them
            continue
        with os.scandir(path_dir) as path:
            for entry in path:
                if entry.is_file() and \
                   _is_backend_filename(entry.name) and \
                   entry.name not in BACKEND_BLACKLIST and \
                   entry.name not in backends:
                    backends.append(entry.name)
    return backends


def start_backends_from_path() -> None:
    """
    Helper for starting all backends found in PATH.
    These backends are expected to have the same command line arguments.
    """

    logger.debug("trying to find backends in PATH")

    for filename in get_backends_from_path():
        logger.debug("found backend file %s in PATH", filename)
        start_backend_from_path(filename)


def start_backend_client(backend: Backend) -> None:
    """
    Helper for starting a single backend client
    """

    logger.debug("starting backend client for backend %s", backend.name)

    # let user know we are connecting
    log_msg = "Starting client for backend \"{0}\".".format(backend.name)
    nuqql.conversation.log_nuqql_conv(log_msg)

    # start backend client and connect to backend server
    backend.start_client()

    # make sure the connection to the backend was successful
    if not backend.client or not backend.client.sock:
        log_msg = "Could not connect to backend \"{0}\".".format(backend.name)
        logger.error("could not connect to backend %s", backend.name)
        nuqql.conversation.log_nuqql_conv(log_msg)
        backend.stop()
        return

    # request accounts from backend
    if not BACKEND_PUSH_ACCOUNTS:
        backend.client.send_accounts()

        # log it
        log_msg = "Collecting accounts for \"{0}\".".format(backend.name)
        if backend.conversation:
            backend.conversation.log("nuqql", log_msg)


def start_backend_clients() -> None:
    """
    Helper for starting all backend clients
    """

    logger.debug("starting backend clients")

    # give backend servers some time
    time.sleep(BACKENDS_WAIT_TIME)

    for backend in dict(BACKENDS).values():
        start_backend_client(backend)


def restart_backend(backend_name: str) -> None:
    """
    (Re)start a backend and the client connection. Called from NuqqlBackend
    """

    logger.debug("restarting backend %s", backend_name)

    if backend_name in BACKENDS:
        # backend already running
        return

    # try to start backend
    backend = None

    # backends are started from PATH
    for filename in get_backends_from_path():
        if backend_name == _get_backend_name(filename):
            backend = start_backend_from_path(filename)
            break

    # start the backend client
    if backend:
        # give backend server some time
        time.sleep(BACKENDS_WAIT_TIME)
        start_backend_client(backend)


def start_nuqql() -> None:
    """
    Start the nuqql dummy backend
    """

    logger.debug("starting nuqql backend")

    # create backend
    backend = NuqqlBackend("nuqql")
    backend.version = VERSION
    backend.backends = BACKENDS
    backend.restart_func = restart_backend

    # add conversation and show it in list window
    conv = nuqql.conversation.NuqqlConversation(backend, None, backend.name)
    conv.create_windows()
    backend.conversation = conv
    conv.wins.list_win.redraw()
    nuqql.conversation.log_nuqql_conv(f"Started nuqql v{VERSION}.")
    logger.debug("started nuqql backend version %s", VERSION)


def start_backends() -> None:
    """
    Helper for starting all backends
    """

    logger.debug("starting all available backends")

    # start nuqql dummy backend
    start_nuqql()

    # start backends
    nuqql.conversation.log_nuqql_conv("Starting backends.")
    start_backends_from_path()

    # start backend clients
    start_backend_clients()


def stop_backends() -> None:
    """
    Helper for stopping all backends
    """

    logger.debug("stopping all backends")
    for backend in dict(BACKENDS).values():
        backend.stop()  # changes BACKENDS



#!/usr/bin/env python3

"""
nuqql setup file
"""

import os
import re
import codecs

from setuptools import find_packages, setup

# setup parameters
DESCRIPTION = "Command line instant messenger inspired by centericq/centerim"
with open("README.md", 'r') as f:
    LONG_DESCRIPTION = f.read()
CLASSIFIERS = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
]


# setup helpers
def read(*parts):
    """
    Read encoded file
    """

    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, *parts), 'r') as enc_file:
        return enc_file.read()


def find_version(*file_paths):
    """
    Find version in encoded file
    """

    version_file = read(*file_paths)
    version_pattern = r"^VERSION = ['\"]([^'\"]*)['\"]"
    version_match = re.search(version_pattern, version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# run setup
setup(
    name="nuqql",
    version=find_version("nuqql", "__init__.py"),
    description=DESCRIPTION,
    license="MIT",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="hwipl",
    author_email="nuqql@hwipl.net",
    url="https://github.com/hwipl/nuqql",
    packages=find_packages(include=["nuqql", "nuqql.*"]),
    entry_points={
        "console_scripts": ["nuqql = nuqql.main:run",
                            "nuqql-keys = nuqql.tools.nuqql_keys:main"]
    },
    classifiers=CLASSIFIERS,
    python_requires='>=3.6',
)



"""
special nuqql main conversation for welcome screen
"""

import logging

import nuqql.config

from .conversation import Conversation

logger = logging.getLogger(__name__)


class MainConversation(Conversation):
    """
    Class for the main/welcome screen conversation
    """

    def create_windows(self) -> None:
        """
        Create windows for this conversation
        """

        logger.debug("creating windows for conversation %s", self.name)

        # create command windows for nuqql
        list_config = nuqql.config.get("list_win")
        log_config = nuqql.config.get("log_win_main")
        self.wins.list_win = nuqql.win.ListWin(list_config, self,
                                               "Conversation list")
        self.wins.log_win = nuqql.win.LogWin(log_config, self,
                                             f"nuqql v{nuqql.VERSION}")

        # list conversations in list window
        self.wins.list_win.list = nuqql.conversation.CONVERSATIONS

        # mark nuqql's list window as active, so main loop does not quit
        self.wins.list_win.state.active = True

        # draw list
        self.wins.list_win.redraw()
        self.wins.log_win.redraw()

        # save windows
        nuqql.win.MAIN_WINS["list"] = self.wins.list_win
        nuqql.win.MAIN_WINS["log"] = self.wins.log_win



"""
Backend server
"""

import logging
import subprocess
import threading

from pathlib import Path
from typing import Optional

logger = logging.getLogger(__name__)

# enable/disable logging of subprocess output
SUBPROCESS_LOGGING = True


class BackendServer:
    """
    Class for a backend's server process
    """

    def __init__(self, cmd: str = "", path: str = "") -> None:
        # server
        self.proc: Optional[subprocess.Popen] = None
        self.server_path = path
        self.server_cmd = cmd

        # subprocess output logging thread
        self.output_logger: Optional[threading.Thread] = None
        self.stop_logger = threading.Event()

    def _start_logging(self) -> None:
        """
        Start logging the backend server's output to the nuqql log
        """

        name = self.server_cmd.split()[0]

        def proc_logger() -> None:
            assert self.proc and self.proc.stdout
            while not self.stop_logger.is_set():
                try:
                    for line in iter(self.proc.stdout.readline, b''):
                        logger.debug("got backend subprocess %s output:\n %s",
                                     name, line.decode())
                except OSError:
                    logger.error("error reading subprocess %s output (read)",
                                 name)
                    return
        self.output_logger = threading.Thread(target=proc_logger)
        self.output_logger.start()

    def start(self) -> None:
        """
        Start the backend's server process
        """

        logger.debug("starting server")

        # make sure server's working directory exists
        Path(self.server_path).mkdir(parents=True, exist_ok=True)

        # if logging is enabled for subprocess use pipes
        stdout = subprocess.DEVNULL
        stderr = subprocess.DEVNULL
        if SUBPROCESS_LOGGING:
            stdout = subprocess.PIPE
            stderr = subprocess.STDOUT

        # start server process
        self.proc = subprocess.Popen(
            self.server_cmd,
            shell=True,
            stdout=stdout,
            stderr=stderr,
            start_new_session=True,     # dont send SIGINT from nuqql to
                                        # subprocess
        )

        if SUBPROCESS_LOGGING:
            self._start_logging()

    def stop(self) -> None:
        """
        Stop the backend's server process
        """

        logger.debug("stopping server")

        # stop running server
        if self.proc:
            self.proc.terminate()

        # stop output logger
        if self.output_logger:
            self.stop_logger.set()
            self.output_logger.join()



"""
special nuqql conversation
"""

import logging

from typing import Tuple

import nuqql.config

from .conversation import Conversation, CONVERSATIONS

logger = logging.getLogger(__name__)


class NuqqlConversation(Conversation):
    """
    Class for the nuqql conversation
    """

    def create_windows(self) -> None:
        """
        Create windows for this conversation
        """

        logger.debug("creating windows for conversation %s", self.name)

        # create command windows for nuqql
        log_title = "Command log of nuqql"
        input_title = "Command to nuqql"
        self._create_windows_common(log_title, input_title)

        # nuqql itself needs a list window for buddy list
        self.wins.list_win = nuqql.win.MAIN_WINS["list"]

        # add to conversation list
        CONVERSATIONS.append(self)

    def get_name(self) -> str:
        """
        Get the name of the conversation, depending on type
        """

        notify = self._get_name_notification()
        return "{0}{{nuqql}}".format(notify)

    def get_key(self) -> Tuple:
        """
        Get a key for sorting this conversation
        """

        # return tuple of sort keys:
        # notify, used, type, status, name
        return 0 - self.notification, 0, 2, 0, self.name

    def send_msg(self, msg: str) -> None:
        """
        Send message coming from the UI/input window
        """

        logger.debug("sending message %s in conversation %s", msg, self.name)

        self._send_msg_prepare(msg)

        # handle nuqql command
        assert isinstance(self.backend, nuqql.backend.NuqqlBackend)
        self.backend.handle_nuqql_command(msg)



"""
History: (file) logging for nuqql conversations
"""

import datetime
import logging
import pathlib
import os

from typing import List, Optional, TYPE_CHECKING

from .logmessage import LogMessage

if TYPE_CHECKING:   # imports for typing
    # pylint: disable=cyclic-import
    from .conversation import Conversation  # noqa

logger = logging.getLogger(__name__)

HISTORY_FILE = "/history"
LASTREAD_FILE = "/lastread"


class History:
    """
    Message history
    """

    def __init__(self, conv: "Conversation") -> None:
        self.conv = conv
        self.conv_path = ""
        self.log: List[LogMessage] = []
        self.logger: Optional[logging.Logger] = None
        self.log_file = ""
        # self.lastread_file = None

    def _get_conv_path(self) -> str:
        """
        Get path for conversation history as a string and make sure it exists
        """

        # construct directory path
        assert self.conv.backend and self.conv.account
        conv_dir = str(pathlib.Path.home()) + \
            "/.config/nuqql/conversation/{}/{}/{}".format(
                self.conv.backend.name, self.conv.account.aid, self.conv.name)

        # make sure directory exists
        pathlib.Path(conv_dir).mkdir(parents=True, exist_ok=True)

        return conv_dir

    @staticmethod
    def _get_logger(name, file_name: str) -> logging.Logger:
        """
        Create a logger for a conversation
        """

        # create logger
        conv_logger = logging.getLogger(name)
        conv_logger.propagate = False
        conv_logger.setLevel(logging.DEBUG)

        # create handler
        fileh = logging.FileHandler(file_name)
        fileh.setLevel(logging.DEBUG)
        fileh.terminator = "\r\n"

        # create formatter
        formatter = logging.Formatter(
            # fmt="%(asctime)s %(levelname)-5.5s [%(name)s] %(message)s",
            fmt="%(message)s",
            datefmt="%s")

        # add formatter to handler
        fileh.setFormatter(formatter)

        # add handler to logger
        if not conv_logger.hasHandlers():
            conv_logger.addHandler(fileh)

        # return logger to caller
        return conv_logger

    def init_logger(self) -> None:
        """
        Init logger for a conversation
        """

        logger.debug("initializing logger of conversation %s", self.conv.name)

        # get log dir and make sure it exists
        assert self.conv.backend and self.conv.account
        self.conv_path = self._get_conv_path()

        # create logger with log name and log file
        log_name = "nuqql.history.{}.{}.{}".format(self.conv.backend.name,
                                                   self.conv.account.aid,
                                                   self.conv.name)
        self.log_file = self.conv_path + HISTORY_FILE
        self.logger = self._get_logger(log_name, self.log_file)

    @staticmethod
    def _parse_log_line(line: str) -> LogMessage:
        """
        Parse line from log file and return a LogMessage
        """

        # parse line
        parts = line.split(sep=" ", maxsplit=3)
        # tstamp = parts[0]
        direction = parts[1]
        is_own = False
        if direction == "OUT":
            is_own = True
        sender = parts[2]
        msg = parts[3][:-2]
        tstamp = datetime.datetime.fromtimestamp(int(parts[0]))

        # create and return LogMessage
        log_msg = LogMessage(tstamp, sender, msg, own=is_own)
        return log_msg

    @staticmethod
    def _create_log_line(log_msg: LogMessage) -> str:
        """
        Create a line for the log files from a LogMessage
        """

        # determine log line contents
        tstamp = round(log_msg.tstamp.timestamp())
        direction = "IN"
        sender = log_msg.sender
        if log_msg.own:
            direction = "OUT"
            sender = "you"
        msg = log_msg.msg

        return "{} {} {} {}".format(tstamp, direction, sender, msg)

    def get_lastread(self) -> Optional[LogMessage]:
        """
        Get last read message from "lastread" file of the conversation
        """

        logger.debug("getting lastread of conversation %s", self.conv.name)
        lastread_file = self.conv_path + LASTREAD_FILE
        try:
            with open(lastread_file, newline="\r\n") as in_file:
                for line in in_file:
                    log_msg = self._parse_log_line(line)
                    log_msg.is_read = True
                    return log_msg
                logger.debug("lastread file of conversation %s is empty",
                             self.conv.name)
                return None
        except FileNotFoundError:
            logger.debug("lastread file of conversation %s not found",
                         self.conv.name)
            return None

    def set_lastread(self, log_msg: LogMessage) -> None:
        """
        Set last read message in "lastread" file of the conversation
        """

        logger.debug("setting lastread of conversation %s", self.conv.name)

        # create log line and write it to lastread file
        line = self._create_log_line(log_msg) + "\r\n"
        lines = []
        lines.append(line)

        lastread_file = self.conv_path + LASTREAD_FILE
        with open(lastread_file, "w+") as out_file:
            out_file.writelines(lines)

    def get_last_log_line(self) -> Optional[LogMessage]:
        """
        Read last LogMessage from log file
        """

        logger.debug("getting last log line of conversation %s",
                     self.conv.name)
        history_file = self.conv_path + HISTORY_FILE
        try:
            # negative seeking requires binary mode
            with open(history_file, "rb") as in_file:
                # check if file contains at least 2 bytes
                in_file.seek(0, os.SEEK_END)
                if in_file.tell() < 3:
                    logger.debug("log of conversation %s seems to be empty",
                                 self.conv.name)
                    return None

                # try to find last line
                in_file.seek(-3, os.SEEK_END)
                while in_file.read(2) != b"\r\n":
                    try:
                        in_file.seek(-3, os.SEEK_CUR)
                    except IOError:
                        in_file.seek(-2, os.SEEK_CUR)
                        if in_file.tell() == 0:
                            break

                # read and return last line as LogMessage
                last_line = in_file.read()
                log_msg = self._parse_log_line(last_line.decode())
                return log_msg
        except FileNotFoundError:
            logger.debug("log file of conversation %s not found",
                         self.conv.name)
            return None

    def init_log_from_file(self) -> None:
        """
        Initialize a conversation's log from the conversation's log file
        """

        logger.debug("initializing log of conversation %s from file %s",
                     self.conv.name, self.log_file)

        # get last read log message
        last_read = self.get_lastread()
        is_read = True

        with open(self.log_file, newline="\r\n") as in_file:
            prev_msg = None
            for line in in_file:
                # parse log line and create log message
                log_msg = self._parse_log_line(line)
                log_msg.is_read = is_read

                # check if date changed between two messages, and print event
                if prev_msg and \
                   prev_msg.tstamp.date() != log_msg.tstamp.date():
                    date_change_msg = LogMessage(
                        log_msg.tstamp,
                        "<event>", "<Date changed to {}>".format(
                            log_msg.tstamp.date()), own=True)
                    date_change_msg.is_read = True
                    self.log.append(date_change_msg)
                prev_msg = log_msg

                # add log message to the conversation's log
                self.log.append(log_msg)

                # if this is the last read message, following message will be
                # marked as unread
                if last_read and last_read.is_equal(log_msg):
                    is_read = False
        if self.log:
            # if there were any log messages in the log file, put a marker in
            # the log where the new messages start
            tstamp = datetime.datetime.now()
            new_conv_msg = "<Started new conversation at {}.>".format(
                tstamp.strftime("%Y-%m-%d %H:%M:%S"))
            log_msg = LogMessage(tstamp, "<event>", new_conv_msg, own=True)
            log_msg.is_read = True
            self.log.append(log_msg)

    def log_to_file(self, log_msg: LogMessage) -> None:
        """
        Write LogMessage to history log file and set lastread message
        """

        logger.debug("logging msg to log file of conversation %s: %s",
                     self.conv.name, log_msg)
        # create line and write it to history
        line = self._create_log_line(log_msg)
        assert self.logger
        self.logger.info(line)

        # assume user read all previous messages when user sends a message and
        # set lastread accordingly
        if log_msg.own:
            self.set_lastread(log_msg)
